# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
General utility functions
"""

from collections import defaultdict
from functools import wraps
from math import floor
from time import time
from typing import Any, Dict, Iterable, List, Literal, Optional, Sequence, Tuple, Union, cast

from more_itertools import chunked
from mthree.utils import final_measurement_mapping
import numpy as np
from qiskit import ClassicalRegister, QuantumCircuit, transpile
from qiskit.converters import circuit_to_dag
from qiskit.transpiler import CouplingMap
import xarray as xr

from iqm.benchmarks.logging_config import qcvv_logger
from iqm.qiskit_iqm import transpile_to_IQM
from iqm.qiskit_iqm.fake_backends.fake_adonis import IQMFakeAdonis
from iqm.qiskit_iqm.fake_backends.fake_apollo import IQMFakeApollo
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from iqm.qiskit_iqm.iqm_job import IQMJob
from iqm.qiskit_iqm.iqm_provider import IQMProvider
from iqm.qiskit_iqm.iqm_transpilation import optimize_single_qubit_gates


def timeit(f):
    """Calculates the amount of time a function takes to execute

    Args:
        f: The function to add the timing attribute to
    Returns:
        The decorated function execution with logger statement of elapsed time in execution
    """

    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        elapsed = te - ts
        if 1.0 <= elapsed <= 60.0:
            qcvv_logger.debug(f'\t"{f.__name__}" took {elapsed:.2f} sec')
        else:
            qcvv_logger.debug(f'\t"{f.__name__}" took {elapsed/60.0:.2f} min')
        return result, elapsed

    return wrap


@timeit
def count_2q_layers(circuit_list: List[QuantumCircuit]) -> List[int]:
    """Calculate the number of layers of parallel 2-qubit gates in a list of circuits.

    Args:
        circuit_list (List[QuantumCircuit]): the list of quantum circuits to analyze.

    Returns:
        List[int]: the number of layers of parallel 2-qubit gates in the list of circuits.
    """
    all_number_2q_layers = []
    for circuit in circuit_list:
        dag = circuit_to_dag(circuit)
        layers = list(dag.layers())  # Call the method and convert the result to a list
        parallel_2q_layers = 0

        for layer in layers:
            two_qubit_gates_in_layer = [
                node
                for node in layer["graph"].op_nodes()  # Use op_nodes to get only operation nodes
                if node.op.num_qubits == 2
            ]
            if two_qubit_gates_in_layer:
                parallel_2q_layers += 1
        all_number_2q_layers.append(parallel_2q_layers)

    return all_number_2q_layers


def count_native_gates(
    backend_arg: Union[str, IQMBackendBase], transpiled_qc_list: List[QuantumCircuit]
) -> Dict[str, Dict[str, float]]:
    """Count the number of IQM native gates of each quantum circuit in a list.

    Args:
        backend_arg (str | IQMBackendBase): The backend, either specified as str or as IQMBackendBase.
        transpiled_qc_list: a list of quantum circuits transpiled to ['r','cz','barrier','measure'] gate set.
    Returns:
        Dictionary with
             - outermost keys being native operations.
             - values being Dict[str, float] with mean and standard deviation values of native operation counts.

    """
    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    native_operations = backend.operation_names
    # Some backends may not include "barrier" in the operation_names attribute
    if "barrier" not in native_operations:
        native_operations.append("barrier")

    num_native_operations: Dict[str, List[int]] = {x: [0] for x in native_operations}
    avg_native_operations: Dict[str, Dict[str, float]] = {x: {} for x in native_operations}

    for q in transpiled_qc_list:
        for k in q.count_ops().keys():
            if k not in native_operations:
                raise ValueError(f"Count # of gates: '{k}' is not in the backend's native gate set")
        for op in native_operations:
            if op in q.count_ops().keys():
                num_native_operations[op].append(q.count_ops()[op])

    avg_native_operations.update(
        {
            x: {"Mean": np.mean(num_native_operations[x]), "Std": np.std(num_native_operations[x])}
            for x in native_operations
        }
    )

    return avg_native_operations


# DD code to be adapted to Pulla version once released
# @timeit
# def execute_with_dd(
#     backend: IQMBackendBase, transpiled_circuits: List[QuantumCircuit], shots: int, dd_strategy: DDStrategy
# ) -> List[Dict[str, int]]:
#     """Executes a list of transpiled quantum circuits with dynamical decoupling according to a specified strategy
#     Args:
#         backend (IQMBackendBase):
#         transpiled_circuits (List[QuantumCircuit]):
#         shots (int):
#         dd_strategy (DDStrategy):
#
#     Returns:
#         List[Dict[str, int]]: The counts of the execution with dynamical decoupling
#     """
#     warnings.warn("Suppressing INFO messages from Pulla with logging.disable(sys.maxsize) - update if problematic!")
#     logging.disable(sys.maxsize)
#
#     pulla_obj = Pulla(cocos_url=iqm_url)
#
#     execution_results = dd.execute_with_dd(
#         pulla_obj,
#         backend=backend,
#         circuits=transpiled_circuits,
#         shots=shots,
#         dd_strategy=dd_strategy,
#     )
#
#     return execution_results


# pylint: disable=too-many-branches
def get_iqm_backend(backend_label: str) -> IQMBackendBase:
    """Get the IQM backend object from a backend name (str).

    Args:
        backend_label (str): The name of the IQM backend.
    Returns:
        IQMBackendBase.
    """
    # ****** 5Q star ******
    # FakeAdonis
    if backend_label.lower() in ("iqmfakeadonis", "fakeadonis"):
        backend_object = IQMFakeAdonis()

    # ****** 20Q grid ******
    # Garnet
    elif backend_label.lower() == "garnet":
        iqm_server_url = "https://cocos.resonance.meetiqm.com/garnet"
        provider = IQMProvider(iqm_server_url)
        backend_object = provider.get_backend()
    # FakeApollo
    elif backend_label.lower() in ("iqmfakeapollo", "fakeapollo"):
        backend_object = IQMFakeApollo()

    # ****** 6Q Resonator Star ******
    # Deneb
    elif backend_label.lower() == "deneb":
        iqm_server_url = "https://cocos.resonance.meetiqm.com/deneb"
        provider = IQMProvider(iqm_server_url)
        backend_object = provider.get_backend()

    else:
        raise ValueError(f"Backend {backend_label} not supported. Try 'garnet', 'deneb', 'fakeadonis' or 'fakeapollo'.")

    return backend_object


def marginal_distribution(prob_dist: Dict[str, float], indices: Iterable[int]) -> Dict[str, float]:
    """Compute the marginal distribution over specified bits (indices)

    Params:
    - prob_dist (dict): A dictionary with keys being bitstrings and values are their probabilities
    - indices (list): List of bit indices to marginalize over

    Returns:
    - dict: A dictionary representing the marginal distribution over the specified bits.
    """
    marginal_dist: Dict[str, float] = defaultdict(float)

    for bitstring, prob in prob_dist.items():
        # Extract the bits at the specified indices and form the marginalized bitstring
        marginalized_bitstring = "".join(bitstring[i] for i in indices)
        # Sum up probabilities for each marginalized bitstring
        marginal_dist[marginalized_bitstring] += prob

    return dict(marginal_dist)


@timeit
def perform_backend_transpilation(
    qc_list: List[QuantumCircuit],
    backend: IQMBackendBase,
    qubits: Sequence[int],
    coupling_map: List[List[int]],
    basis_gates: Tuple[str, ...] = ("r", "cz"),
    qiskit_optim_level: int = 1,
    optimize_sqg: bool = False,
    drop_final_rz: bool = True,
    routing_method: Optional[str] = "sabre",
) -> List[QuantumCircuit]:
    """
    Transpile a list of circuits to backend specifications.

    Args:
        qc_list (List[QuantumCircuit]): The original (untranspiled) list of quantum circuits.
        backend (IQMBackendBase ): The backend to execute the benchmark on.
        qubits (Sequence[int]): The qubits to target in the transpilation.
        coupling_map (List[List[int]]): The target coupling map to transpile to.
        basis_gates (Tuple[str, ...]): The basis gates.
        qiskit_optim_level (int): Qiskit "optimization_level" value.
        optimize_sqg (bool): Whether SQG optimization is performed taking into account virtual Z.
        drop_final_rz (bool): Whether the SQG optimizer drops a final RZ gate.
        routing_method (Optional[str]): The routing method employed by Qiskit's transpilation pass.

    Returns:
        List[QuantumCircuit]: A list of transpiled quantum circuits.
    """

    # Helper function considering whether optimize_sqg is done,
    # and whether the coupling map is reduced (whether final physical layout must be fixed onto an auxiliary QC)
    def transpile_and_optimize(qc, aux_qc=None):
        transpiled = transpile(
            qc,
            basis_gates=basis_gates,
            coupling_map=coupling_map,
            optimization_level=qiskit_optim_level,
            initial_layout=qubits if aux_qc is None else None,
            routing_method=routing_method,
        )
        if optimize_sqg:
            transpiled = optimize_single_qubit_gates(transpiled, drop_final_rz=drop_final_rz)
        if backend.name == "IQMNdonisBackend":
            transpiled = transpile_to_IQM(
                transpiled, backend=backend, optimize_single_qubits=optimize_sqg, remove_final_rzs=drop_final_rz
            )
        if aux_qc is not None:
            if backend.name == "IQMNdonisBackend":
                transpiled = reduce_to_active_qubits(transpiled, backend.name)
                transpiled = aux_qc.compose(transpiled, qubits=[0] + qubits, clbits=list(range(qc.num_clbits)))
            else:
                transpiled = aux_qc.compose(transpiled, qubits=qubits, clbits=list(range(qc.num_clbits)))

        return transpiled

    qcvv_logger.info(
        f"Transpiling for backend {backend.name} with optimization level {qiskit_optim_level}, "
        f"{routing_method} routing method{' and SQG optimization' if optimize_sqg else ''} all circuits"
    )

    if coupling_map == backend.coupling_map:
        transpiled_qc_list = [transpile_and_optimize(qc) for qc in qc_list]
    else:  # The coupling map will be reduced if the physical layout is to be fixed
        aux_qc_list = [QuantumCircuit(backend.num_qubits, q.num_clbits) for q in qc_list]
        transpiled_qc_list = [transpile_and_optimize(qc, aux_qc=aux_qc_list[idx]) for idx, qc in enumerate(qc_list)]

    return transpiled_qc_list


def reduce_to_active_qubits(circuit: QuantumCircuit, backend_name: Optional[str] = None) -> QuantumCircuit:
    """
    Reduces a quantum circuit to only its active qubits.

    Args:
        backend_name (Optional[str]): The backend name, if any, in which the circuits are defined.
        circuit (QuantumCircuit): The original quantum circuit.

    Returns:
        QuantumCircuit: A new quantum circuit containing only active qubits.
    """
    # Identify active qubits
    active_qubits = set()
    for instruction in circuit.data:
        for qubit in instruction.qubits:
            active_qubits.add(circuit.find_bit(qubit).index)
    if backend_name is not None and backend_name == "IQMNdonisBackend" and 0 not in active_qubits:
        # For star systems, the resonator must always be there, regardless of whether it MOVE gates on it or not
        active_qubits.add(0)

    # Create a mapping from old qubits to new qubits
    active_qubits = set(sorted(active_qubits))
    qubit_map = {old_idx: new_idx for new_idx, old_idx in enumerate(active_qubits)}

    # Create a new quantum circuit with the reduced number of qubits
    reduced_circuit = QuantumCircuit(len(active_qubits))

    # Add classical registers if they exist
    if circuit.num_clbits > 0:
        creg = ClassicalRegister(circuit.num_clbits)
        reduced_circuit.add_register(creg)

    # Copy operations to the new circuit, remapping qubits and classical bits
    for instruction in circuit.data:
        new_qubits = [reduced_circuit.qubits[qubit_map[circuit.find_bit(qubit).index]] for qubit in instruction.qubits]
        new_clbits = [reduced_circuit.clbits[circuit.find_bit(clbit).index] for clbit in instruction.clbits]
        reduced_circuit.append(instruction.operation, new_qubits, new_clbits)

    return reduced_circuit


@timeit
def retrieve_all_counts(iqm_jobs: List[IQMJob], identifier: Optional[str] = None) -> List[Dict[str, int]]:
    """Retrieve the counts from a list of IQMJob objects.
    Args:
        iqm_jobs (List[IQMJob]): The list of IQMJob objects.
        identifier (Optional[str]): a string identifying the job.
    Returns:
        List[Dict[str, int]]: The counts of all the IQMJob objects.
    """
    if identifier is None:
        qcvv_logger.info(f"Retrieving all counts")
    else:
        qcvv_logger.info(f"Retrieving all counts for {identifier}")
    final_counts = []
    for j in iqm_jobs:
        counts = j.result().get_counts()
        if isinstance(counts, list):
            final_counts.extend(counts)
        elif isinstance(counts, dict):
            final_counts.append(counts)

    return final_counts


def retrieve_all_job_metadata(
    iqm_jobs: List[IQMJob],
) -> Dict[str, Dict[str, Any]]:
    """Retrieve the counts from a list of IQMJob objects.
    Args:
        iqm_jobs List[IQMJob]: The list of IQMJob objects.

    Returns:
        Dict[str, Dict[str, Any]]: Relevant metadata of all the IQMJob objects.
    """
    all_meta = {}

    for index, j in enumerate(iqm_jobs):
        all_attributes_j = dir(j)
        all_meta.update(
            {
                "batch_job_"
                + str(index + 1): {
                    "job_id": j.job_id() if "job_id" in all_attributes_j else None,
                    "backend": j.backend().name if "backend" in all_attributes_j else None,
                    "status": j.status().value if "status" in all_attributes_j else None,
                    "circuits_in_batch": (
                        len(cast(List, j.circuit_metadata)) if "circuit_metadata" in all_attributes_j else None
                    ),
                    "shots": j.metadata["shots"] if "shots" in j.metadata.keys() else None,
                    "timestamps": j.metadata["timestamps"] if "timestamps" in j.metadata.keys() else None,
                }
            }
        )

    return all_meta


def set_coupling_map(
    qubits: Sequence[int], backend: IQMBackendBase, physical_layout: Literal["fixed", "batching"] = "fixed"
) -> CouplingMap:
    """Set a coupling map according to the specified physical layout.

    Args:
        qubits (Sequence[int]): the list of physical qubits to consider.
        backend (IQMBackendBase): the backend from IQM.
        physical_layout (Literal["fixed", "batching"]): the physical layout type to consider.
                - "fixed" sets a coupling map restricted to the input qubits -> results will be constrained to measure those qubits.
                - "batching" sets the coupling map of the backend -> results in a benchmark will be "batched" according to final layouts.
                * Default is "fixed".
    Returns:
        A coupling map according to the specified physical layout.
    """
    if physical_layout == "fixed":
        if backend.name == "IQMNdonisBackend":
            return backend.coupling_map.reduce(mapping=[0] + list(qubits))
        return backend.coupling_map.reduce(mapping=qubits)
    if physical_layout == "batching":
        return backend.coupling_map
    raise ValueError('physical_layout must either be "fixed" or "batching"')


@timeit
def sort_batches_by_final_layout(
    transpiled_circuit_list: List[QuantumCircuit],
) -> Tuple[Dict[Tuple, List[QuantumCircuit]], Dict[Tuple, List[int]]]:
    """Sort batches of circuits according to the final measurement mapping in their corresponding backend.

    Args:
        transpiled_circuit_list (List[QuantumCircuit]): the list of circuits transpiled to a given backend.
    Returns:
        sorted_circuits (Dict[Tuple, List[QuantumCircuit]]): dictionary, keys: final measured qubits, values: corresponding circuits.
        sorted_indices (Dict[Tuple, List[int]]): dictionary, keys: final measured qubits, values: corresponding circuit indices.
    """
    qcvv_logger.info("Now getting the final measurement maps of all circuits")
    all_measurement_maps = [tuple(final_measurement_mapping(qc).values()) for qc in transpiled_circuit_list]
    unique_measurement_maps = set(tuple(sorted(x)) for x in all_measurement_maps)
    sorted_circuits: Dict[Tuple, List[QuantumCircuit]] = {u: [] for u in unique_measurement_maps}
    sorted_indices: Dict[Tuple, List[int]] = {i: [] for i in unique_measurement_maps}
    for index, qc in enumerate(transpiled_circuit_list):
        final_measurement = all_measurement_maps[index]
        final_measurement = tuple(sorted(final_measurement))
        sorted_circuits[final_measurement].append(qc)
        sorted_indices[final_measurement].append(index)

    if len(sorted_circuits) == 1:
        qcvv_logger.info(f"The routing method generated a single batch of circuits to be measured")
    else:
        qcvv_logger.info(f"The routing method generated {len(sorted_circuits)} batches of circuits to be measured")

    return sorted_circuits, sorted_indices


@timeit
def submit_execute(
    sorted_transpiled_qc_list: Dict[Tuple, List[QuantumCircuit]],
    backend: IQMBackendBase,
    shots: int,
    calset_id: Optional[str],
    max_gates_per_batch: Optional[int],
) -> List[IQMJob]:
    """Submit for execute a list of quantum circuits on the specified Backend.

    Args:
        sorted_transpiled_qc_list (Dict[Tuple, List[QuantumCircuit]]): the list of quantum circuits to be executed.
        backend (IQMBackendBase): the backend to execute the circuits on.
        shots (int): the number of shots per circuit.
        calset_id (Optional[str]): the calibration set ID, uses the latest one if None.
        max_gates_per_batch (int): the maximum number of gates per batch sent to the backend, used to make manageable batches.
    Returns:
        List[IQMJob]: the IQMJob objects of the executed circuits.
    """
    final_jobs = []
    for k in sorted(
        sorted_transpiled_qc_list.keys(),
        key=lambda x: len(sorted_transpiled_qc_list[x]),
        reverse=True,
    ):
        # sorted is so batches are looped from larger to smaller
        qcvv_logger.info(
            f"Submitting batch with {len(sorted_transpiled_qc_list[k])} circuits corresponding to qubits {list(k)}"
        )
        # Divide into batches according to maximum gate count per batch
        if max_gates_per_batch is None:
            jobs = backend.run(sorted_transpiled_qc_list[k], shots=shots, calibration_set_id=calset_id)
            final_jobs.append(jobs)
        else:
            # Calculate average gate count per quantum circuit
            avg_gates_per_qc = sum(sum(qc.count_ops().values()) for qc in sorted_transpiled_qc_list[k]) / len(
                sorted_transpiled_qc_list[k]
            )
            final_batch_jobs = []
            for index, qc_batch in enumerate(
                chunked(
                    sorted_transpiled_qc_list[k],
                    max(1, floor(max_gates_per_batch / avg_gates_per_qc)),
                )
            ):
                qcvv_logger.info(
                    f"max_gates_per_batch restriction: submitting subbatch #{index+1} with {len(qc_batch)} circuits corresponding to qubits {list(k)}"
                )
                batch_jobs = backend.run(qc_batch, shots=shots, calibration_set_id=calset_id)
                final_batch_jobs.append(batch_jobs)
            final_jobs.extend(final_batch_jobs)

    return final_jobs


def xrvariable_to_counts(dataset: xr.Dataset, identifier: str, counts_range: int) -> List[Dict[str, int]]:
    """Retrieve counts from xarray dataset.

    Args:
        dataset (xr.Dataset): the dataset to extract counts from.
        identifier (str): the identifier for the dataset counts.
        counts_range (int): the range of counts to extract (e.g., the amount of circuits that were executed).
    Returns:
        List[Dict[str, int]]: A list of counts dictionaries from the dataset.
    """
    return [
        dict(zip(list(dataset[f"{identifier}_state_{u}"].data), dataset[f"{identifier}_counts_{u}"].data))
        for u in range(counts_range)
    ]



"""Tests for volumetric benchmarks"""

from iqm.benchmarks.quantum_volume.clops import CLOPSBenchmark, CLOPSConfiguration
from iqm.benchmarks.quantum_volume.quantum_volume import QuantumVolumeBenchmark, QuantumVolumeConfiguration


backend = "fakeapollo"


class TestQV:
    def test_qv(self):
        EXAMPLE_QV = QuantumVolumeConfiguration(
            num_circuits=5,
            shots=2**4,
            calset_id=None,
            num_sigmas=2,
            choose_qubits_routine="custom",
            custom_qubits_array=[[0, 1]],
            qiskit_optim_level=3,
            optimize_sqg=True,
            routing_method="sabre",
            physical_layout="fixed",
            max_gates_per_batch=60_000,
            rem=True,
            mit_shots=10,
        )
        benchmark = QuantumVolumeBenchmark(backend, EXAMPLE_QV)
        benchmark.run()
        benchmark.analyze()

    def test_clops(self):
        EXAMPLE_CLOPS = CLOPSConfiguration(
            qubits=[0, 1],
            num_circuits=2,  # By definition set to 100
            num_updates=2,  # By definition set to 10
            num_shots=2,  # By definition set to 100
            calset_id=None,
            clops_h_bool=True,
            qiskit_optim_level=3,
            optimize_sqg=True,
            routing_method="sabre",
            physical_layout="fixed",
        )
        benchmark = CLOPSBenchmark(backend, EXAMPLE_CLOPS)
        benchmark.run()
        benchmark.analyze()



"""Tests for compressive GST benchmark"""

from unittest.mock import patch

# General BenchmarkExperiment
from iqm.benchmarks.benchmark_experiment import BenchmarkExperiment
from iqm.benchmarks.compressive_gst.compressive_gst import GSTConfiguration


backend = "iqmfakeapollo"


class TestGST:
    @patch('matplotlib.pyplot.figure')
    def test_1q(self, mock_fig):
        Minimal_1Q_GST = GSTConfiguration(
            qubits=[5], gate_set="1QXYI", num_circuits=10, shots=10, rank=4, bootstrap_samples=0, max_iterations=[1, 1]
        )
        EXAMPLE_EXPERIMENT = BenchmarkExperiment(backend, [Minimal_1Q_GST])
        EXAMPLE_EXPERIMENT.run_experiment()
        mock_fig.assert_called()

    @patch('matplotlib.pyplot.figure')
    def test_2q(self, mock_fig):
        Minimal_2Q_GST = GSTConfiguration(
            qubits=[0, 3],
            gate_set="2QXYCZ_extended",
            num_circuits=10,
            shots=10,
            rank=1,
            bootstrap_samples=0,
            max_iterations=[1, 1],
        )
        EXAMPLE_EXPERIMENT = BenchmarkExperiment(backend, [Minimal_2Q_GST])
        EXAMPLE_EXPERIMENT.run_experiment()
        mock_fig.assert_called()



"""Tests for volumetric benchmarks"""

from unittest.mock import patch

from iqm.benchmarks.benchmark_experiment import BenchmarkExperiment
from iqm.benchmarks.optimization.qscore import QScoreConfiguration


backend = "IQMFakeAdonis"


class TestQScore:
    @patch('matplotlib.pyplot.figure')
    def test_qscore(self, mock_fig):
        EXAMPLE_QSCORE = QScoreConfiguration(
            num_instances=2,
            num_qaoa_layers=1,
            shots=4,
            calset_id=None,  # calibration set ID, default is None
            min_num_nodes=2,
            max_num_nodes=5,
            use_virtual_node=True,
            use_classically_optimized_angles=True,
            choose_qubits_routine="custom",
            custom_qubits_array=[[2], [2, 0], [2, 0, 1], [2, 0, 1, 3], [2, 0, 1, 3, 4]],
            seed=1,
        )
        EXAMPLE_EXPERIMENT = BenchmarkExperiment(backend, [EXAMPLE_QSCORE])
        EXAMPLE_EXPERIMENT.run_experiment()
        mock_fig.assert_called()



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Generic Benchmark Experiment class
"""

from copy import deepcopy
from json import dump
from pathlib import Path
import pickle
from time import strftime
from typing import List, Optional, OrderedDict, Union

from iqm.benchmarks.benchmark import BenchmarkBase, BenchmarkConfigurationBase
from iqm.benchmarks.utils import get_iqm_backend
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


class BenchmarkExperiment:
    """
    A Benchmark Experiment wraps the execution of one or more benchmarks, checks their requirements are met, stores the
    execution results and plots relevant figures.
    """

    def __init__(
        self,
        backend: Union[str, IQMBackendBase],
        benchmark_configurations: List[BenchmarkConfigurationBase],
        device_id: Optional[str] = None,
    ):
        """Construct the BenchmarkExperiment class.

        Args:
            backend (str | IQMBackendBase): the backend to execute the benchmarks on
            benchmark_configurations (List[BenchmarkConfigurationBase]): the configuration(s) of the benchmark(s)
            device_id (Optional[str], optional): the identifier of the device. Defaults to None.

        Raises:
            ValueError: backend not supported. Try 'garnet' or 'iqmfakeadonis'
        """
        self.timestamp = strftime("%Y%m%d-%H%M%S")

        self.device_id = device_id if device_id is not None else backend

        if isinstance(backend, str):
            self.backend = get_iqm_backend(backend)

        else:
            assert isinstance(backend, IQMBackendBase)
            self.backend = backend

        benchmarks: OrderedDict[str, BenchmarkBase] = OrderedDict(
            (config.benchmark.name(), config.benchmark(self.backend, config)) for config in benchmark_configurations
        )

        for benchmark in benchmarks.values():
            benchmarks_copy = deepcopy(benchmarks)
            benchmarks_copy = benchmark.check_requirements(benchmarks_copy)
        self.benchmarks = benchmarks_copy

    def run_experiment(self) -> None:
        """Run the Benchmark experiment, and store the configuration, raw data, results and figures."""

        for name, benchmark in self.benchmarks.items():
            print("\nNow executing " + name)
            # Create the directory for results
            results_dir = f"Outputs/{self.device_id}/{self.timestamp}/{name}/"
            Path(results_dir).mkdir(parents=True, exist_ok=True)

            # Execute the current benchmark
            benchmark.generate_requirements(self.benchmarks)
            benchmark.execute_full_benchmark()

            # Create configuration JSON file
            with open(
                f"{results_dir}{self.device_id}_{self.timestamp}_{name}_configuration.json",
                "w",
                encoding="utf-8",
            ) as f_configuration:
                dump(benchmark.serializable_configuration.model_dump(), f_configuration)

            # Create untranspiled circuit files
            if benchmark.untranspiled_circuits:
                for key_qubits in benchmark.untranspiled_circuits.keys():
                    for key_depth in benchmark.untranspiled_circuits[key_qubits].keys():
                        with open(
                            f"{results_dir}{self.device_id}_{self.timestamp}_{name}_qubits_{key_qubits}_depth_{key_depth}_untranspiled.pkl",
                            "wb",
                        ) as f_circuits:
                            pickle.dump(
                                benchmark.untranspiled_circuits[key_qubits][key_depth],
                                f_circuits,
                            )

            # Create transpiled circuit files
            for key_qubits in benchmark.transpiled_circuits.keys():
                for key_depth in benchmark.transpiled_circuits[key_qubits].keys():
                    with open(
                        f"{results_dir}{self.device_id}_{self.timestamp}_{name}_qubits_{key_qubits}_depth_{key_depth}_transpiled.pkl",
                        "wb",
                    ) as f_circuits:
                        pickle.dump(
                            benchmark.transpiled_circuits[key_qubits][key_depth],
                            f_circuits,
                        )

            # Create raw result pickle files
            with open(
                f"{results_dir}{self.device_id}_{self.timestamp}_{name}_raw_results.pkl",
                "wb",
            ) as f_raw_results:
                pickle.dump(
                    benchmark.raw_results,
                    f_raw_results,
                )

            # Create raw data JSON file
            with open(
                f"{results_dir}{self.device_id}_{self.timestamp}_{name}_raw_data.json",
                "w",
                encoding="utf-8",
            ) as f_raw_data:
                dump(benchmark.raw_data, f_raw_data)

            # Create job metadata JSON file
            with open(
                f"{results_dir}{self.device_id}_{self.timestamp}_{name}_job_metadata.json",
                "w",
                encoding="utf-8",
            ) as f_job_metadata:
                dump(benchmark.job_meta, f_job_metadata)

            # Create results JSON file
            with open(
                f"{results_dir}{self.device_id}_{self.timestamp}_{name}_results.json",
                "w",
                encoding="utf-8",
            ) as f_results:
                dump(benchmark.results, f_results)

            # Create figures
            Path(f"{results_dir}figures/").mkdir(parents=True, exist_ok=True)
            for fig_name, fig in benchmark.figures.items():
                fig.savefig(
                    f"{results_dir}figures/{self.device_id}_{self.timestamp}_{name}_{fig_name}",
                    dpi=250,
                    bbox_inches="tight",
                )

            # Save benchmark
            self.benchmarks[benchmark.name()] = benchmark



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""M3 modification for readout mitigation at IQM QPU's."""
import logging
from math import ceil
import sys
import threading
from typing import Any, Dict, Iterable, List
import warnings

import mthree
from mthree.circuits import _marg_meas_states, _tensor_meas_states, balanced_cal_circuits, balanced_cal_strings
from mthree.classes import QuasiCollection
from mthree.exceptions import M3Error
from mthree.mitigation import _job_thread
from mthree.utils import final_measurement_mapping
from qiskit import QuantumCircuit, execute  # pylint: disable = no-name-in-module
from qiskit.providers import Backend, BackendV1, BackendV2

from iqm.benchmarks.utils import get_iqm_backend, timeit
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


# The code here is close to the original M3 code licenced under Apache 2 as well (https://github.com/Qiskit/qiskit-addon-mthree/blob/main/LICENSE.txt).
# We deactivate this to not break functionality.
# pylint: disable=too-many-instance-attributes, too-many-arguments, too-many-locals, too-many-branches, too-many-statements,
# pylint: disable=attribute-defined-outside-init, too-few-public-methods, access-member-before-definition
class M3IQM(mthree.M3Mitigation):
    """M3 readout mitigation class modified to work with IQM devices."""

    def cals_from_system(
        self,
        qubits=None,
        shots=None,
        method=None,
        initial_reset=False,
        rep_delay=None,
        cals_file=None,
        async_cal=False,
        cal_id=None,
    ):
        """Grab calibration data from system.

        Parameters:
            qubits (array_like): Qubits over which to correct calibration data. Default is all.
            shots (int): Number of shots per circuit. min(1e4, max_shots).
            method (str): Type of calibration, 'balanced' (default for hardware),
                         'independent' (default for simulators), or 'marginal'.
            initial_reset (bool): Use resets at beginning of calibration circuits, default=False.
            rep_delay (float): Delay between circuits on IBM Quantum backends.
            cals_file (str): Output path to write JSON calibration data to.
            async_cal (bool): Do calibration async in a separate thread, default is False.

        Raises:
            M3Error: Called while a calibration currently in progress.
        """
        if self._thread:
            raise M3Error("Calibration currently in progress.")
        if qubits is None:
            qubits = range(self.num_qubits)
            # Remove faulty qubits if any
            if any(self.system_info["inoperable_qubits"]):
                qubits = list(
                    filter(
                        lambda item: item not in self.system_info["inoperable_qubits"],
                        list(range(self.num_qubits)),
                    )
                )
                warnings.warn(
                    f"Backend reporting inoperable qubits. Skipping calibrations for: %s",
                    self.system_info["inoperable_qubits"],
                )

        if method is None:
            method = "balanced"
            # if self.system_info["simulator"]:
            #     method = "independent"
        self.cal_method = method
        self.rep_delay = rep_delay
        self.cals_file = cals_file
        self.cal_timestamp = None
        self._grab_additional_cals(
            qubits,
            shots=shots,
            method=method,
            rep_delay=rep_delay,
            initial_reset=initial_reset,
            async_cal=async_cal,
            cal_id=cal_id,
        )

    def _grab_additional_cals(
        self,
        qubits,
        shots=None,
        method="balanced",
        rep_delay=None,
        initial_reset=False,
        async_cal=False,
        cal_id=None,
    ):
        """Grab missing calibration data from backend.

        Parameters:
            qubits (array_like): List of measured qubits.
            shots (int): Number of shots to take, min(1e4, max_shots).
            method (str): Type of calibration, 'balanced' (default), 'independent', or 'marginal'.
            rep_delay (float): Delay between circuits on IBM Quantum backends.
            initial_reset (bool): Use resets at beginning of calibration circuits, default=False.
            async_cal (bool): Do calibration async in a separate thread, default is False.

        Raises:
            M3Error: Backend not set.
            M3Error: Faulty qubits found.
        """
        if self.system is None:
            raise M3Error("System is not set.  Use 'cals_from_file'.")
        if self.single_qubit_cals is None:
            self.single_qubit_cals = [None] * self.num_qubits
        if self.cal_shots is None:
            if shots is None:
                shots = min(self.system_info["max_shots"], 10000)
            self.cal_shots = shots
        if self.rep_delay is None:
            self.rep_delay = rep_delay

        if method not in ["independent", "balanced", "marginal"]:
            raise M3Error(
                f"Invalid calibration method: {method}. Valid methods are 'independent', 'balanced', or 'marginal'."
            )

        if isinstance(qubits, dict):
            # Assuming passed a mapping
            qubits = list(set(qubits.values()))
        elif isinstance(qubits, list):
            # Check if passed a list of mappings
            if isinstance(qubits[0], dict):
                # Assuming list of mappings, need to get unique elements
                _qubits = []
                for item in qubits:
                    _qubits.extend(list(set(item.values())))
                qubits = list(set(_qubits))

        # Do check for inoperable qubits here
        inoperable_overlap = list(set(qubits) & set(self.system_info["inoperable_qubits"]))
        if any(inoperable_overlap):
            raise M3Error(f"Attempting to calibrate inoperable qubits: {inoperable_overlap}")

        num_cal_qubits = len(qubits)
        cal_strings = []
        # shots is needed here because balanced cals will use a value
        # different from cal_shots
        shots = self.cal_shots
        if method == "marginal":
            trans_qcs = _marg_meas_states(qubits, self.num_qubits, initial_reset=initial_reset)
        elif method == "balanced":
            cal_strings = balanced_cal_strings(num_cal_qubits)
            trans_qcs = balanced_cal_circuits(cal_strings, qubits, self.num_qubits, initial_reset=initial_reset)
            shots = self.cal_shots // num_cal_qubits
            if self.cal_shots / num_cal_qubits != shots:
                shots += 1
            self._balanced_shots = shots * num_cal_qubits
        # Independent
        else:
            trans_qcs = []
            for qubit in qubits:
                trans_qcs.extend(_tensor_meas_states(qubit, self.num_qubits, initial_reset=initial_reset))

        num_circs = len(trans_qcs)
        # check for max number of circuits per job
        if isinstance(self.system, BackendV1):
            # Aer simulator has no 'max_experiments'
            max_circuits = getattr(self.system.configuration(), "max_experiments", 300)
        elif isinstance(self.system, BackendV2):
            max_circuits = self.system.max_circuits
            # Needed for https://github.com/Qiskit/qiskit-terra/issues/9947
            if max_circuits is None:
                max_circuits = 300
        else:
            raise M3Error("Unknown backend type")
        # Determine the number of jobs required
        num_jobs = ceil(num_circs / max_circuits)
        # Get the slice length
        circ_slice = ceil(num_circs / num_jobs)
        circs_list = [trans_qcs[kk * circ_slice : (kk + 1) * circ_slice] for kk in range(num_jobs - 1)] + [
            trans_qcs[(num_jobs - 1) * circ_slice :]
        ]
        # Do job submission here
        # This Backend check is here for Qiskit direct access.  Should be removed later.
        jobs = []
        if not isinstance(self.system, Backend):
            for circs in circs_list:
                _job = execute(
                    circs,
                    self.system,
                    optimization_level=0,
                    shots=shots,
                    rep_delay=self.rep_delay,
                )
                jobs.append(_job)

        # *****************************************
        else:  # That's what IQM backend do!
            # *****************************************
            for circs in circs_list:
                if cal_id is None:
                    _job = self.system.run(circs, shots=shots, rep_delay=self.rep_delay)
                else:
                    _job = self.system.run(
                        circs,
                        shots=shots,
                        rep_delay=self.rep_delay,
                        calibration_set_id=cal_id,
                    )
                jobs.append(_job)
                print(f"{len(circs)} calibration circuits ready to be executed!")

        # Execute job and cal building in new thread.
        self._job_error = None
        if async_cal:
            thread = threading.Thread(
                target=_job_thread,
                args=(jobs, self, qubits, num_cal_qubits, cal_strings),
            )
            self._thread = thread
            self._thread.start()
        else:
            _job_thread(jobs, self, qubits, num_cal_qubits, cal_strings)


def readout_error_m3(counts: Dict[str, float], mit: M3IQM, qubits: Iterable) -> Dict[str, float]:
    """Counts processor using M3IQM for readout error mitigation.

    The qubits argument can be either a dictionary coming from `mthree.utils.final_measurement_mapping`
    or an array like the initial layout coming from [backend.qubit_name_to_index(name)]

    returns a dictionary of quasiprobabilties.

    NOTE: we could also pass a list of input counts and then this would return a list of quasiprobabilities.
    This would not work out of the box for us since we need the annotations of either Dict or List (not Union).
    """
    return mit.apply_correction(counts, qubits)


@timeit
def apply_readout_error_mitigation(
    backend_arg: str | IQMBackendBase,
    transpiled_circuits: List[QuantumCircuit],
    counts: List[Dict[str, int]],
    mit_shots: int = 1000,
) -> List[tuple[Any, Any]] | List[tuple[QuasiCollection, list]] | List[QuasiCollection]:
    """
    Args:
        backend_arg (str | IQMBackendBase): the backend to calibrate an M3 mitigator against.
        transpiled_circuits (List[QuantumCircuit]): the list of transpiled quantum circuits.
        counts (List[Dict[str, int]]): the measurement counts corresponding to the circuits.
        mit_shots (int): number of shots per circuit.
    Returns:
        tuple[Any, Any] | tuple[QuasiCollection, list] | QuasiCollection: a list of dictionaries with REM-corrected quasiprobabilities for each outcome.
    """
    # M3IQM uses mthree.mitigation, which for some reason displays many INFO messages
    # Not sure if this is the best way to suppress them; MODIFY IF NECESSARY !
    warnings.warn("Suppressing INFO messages from M3IQM with logging.disable(sys.maxsize) - update if problematic!")
    logging.disable(sys.maxsize)

    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    # Initialize with the given system and get calibration data
    qubits_rem = [final_measurement_mapping(c) for c in transpiled_circuits]

    mit = M3IQM(backend)
    mit.cals_from_system(qubits_rem, shots=mit_shots)
    # Apply the REM correction to the given measured counts
    rem_quasidistro = [mit.apply_correction(c, q) for c, q in zip(counts, qubits_rem)]

    return rem_quasidistro



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements the base class for defining a benchmark.
"""

from abc import ABC, abstractmethod
import copy
from copy import deepcopy
from dataclasses import dataclass, field
import functools
from typing import Any, Dict, List, Optional, Union
import uuid

from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import xarray as xr

from iqm.benchmarks.benchmark import BenchmarkConfigurationBase
from iqm.benchmarks.utils import get_iqm_backend, timeit
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase
from iqm.qiskit_iqm.iqm_provider import IQMBackend, IQMFacadeBackend


@dataclass
class BenchmarkObservationIdentifier:
    """Identifier for observations for ease of use

    Attributes:
        qubit_indices: list containing the indices of the qubits the observation was executed on.
    """

    qubit_indices: list[int]

    @property
    def string_identifier(self) -> str:
        """String version of the qubit indices for ease of use

        Returns:
            A string of the qubit indices
        """
        return str(self.qubit_indices)


@dataclass
class BenchmarkObservation:
    """Dataclass to store the main results of a single run of a Benchmark

    Attributes:
        name: name of the observation
        value: value of the observation
        identifier: identifier, which should be a string of the qubit layout
        uncertainty: uncertainty of the observation
    """

    name: str
    value: Any
    identifier: BenchmarkObservationIdentifier
    uncertainty: Optional[Any] = None


@dataclass
class BenchmarkRunResult:
    """
    A dataclass that stores the results of a single run of a Benchmark.

    RunResult should contain enough information that the Benchmark can be analyzed based on those
    results.
    """

    dataset: xr.Dataset


@dataclass
class BenchmarkAnalysisResult:
    """
    A dataclass storing the results of the analysis.

    The result consists of a dataset, plots, and observations. Plots are defined as a dictionary
    that maps a plot name to a figure. Observations are key-value pairs of data that contain the
    main results of the benchmark.
    """

    dataset: xr.Dataset
    plots: dict[str, Figure] = field(default_factory=lambda: ({}))
    observations: list[BenchmarkObservation] = field(default_factory=lambda: [])

    def plot(self, plot_name: str):
        """
        Plots the given figure.

        Args:
            plot_name: Name of the figure to be plotted.
        """
        return self.plots[plot_name]

    def plot_all(self):
        """
        Plots all the figures defined for the analysis.
        """
        for fig in self.plots.values():
            show_figure(fig)
            plt.show()

    @classmethod
    def from_run_result(cls, run: BenchmarkRunResult):
        """
        Creates a new ``AnalysisResult`` from a ``RunResult``.

        Args:
            run: A run for which analysis result is created.
        """
        return cls(dataset=run.dataset)


def default_analysis_function(result: BenchmarkAnalysisResult) -> BenchmarkAnalysisResult:
    """
    The default analysis that only pass the result through.
    """
    return result


def merge_datasets_dac(datasets: List[xr.Dataset]) -> xr.Dataset:
    """
    Merges a list of datasets recursively to minimize dataset sizes during merge
    Args:
        datasets: List[xr.Dateset]
            A list of xarray datasets
    Returns:
        A list containing a single merged dataset
    """
    if len(datasets) == 1:
        return datasets[0]
    datasets_new = []
    for i in range(0, len(datasets), 2):
        if i == (len(datasets) - 1):
            datasets_new.append(datasets[i])
        else:
            datasets_new.append(xr.merge(datasets[i : i + 2]))
    return merge_datasets_dac(datasets_new)


@timeit
def add_counts_to_dataset(counts: List[Dict[str, int]], identifier: str, dataset: xr.Dataset):
    """Adds the counts from a cortex job result to the given dataset.
    If counts with the same identifier are already present in the old dataset, then both counts are added together.

    Args:
        counts (List[Dict[str, int]]): A list of dictionaries with counts of bitstrings.
        identifier (str): A string to identify the current data, for instance the qubit layout.
        dataset (xr.Dataset): Dataset to add results to.
    Returns:
        dataset_merged: xarray.Dataset
            A merged dataset where the new counts are added the input dataset
    """
    if not isinstance(counts, list):
        counts = [counts]
    datasets = []
    for ii, _ in enumerate(counts):
        ds_temp = xr.Dataset()
        counts_dict = dict(counts[ii])
        counts_array = xr.DataArray(list(counts_dict.values()), {f"{identifier}_state_{ii}": list(counts_dict.keys())})
        if f"{identifier}_counts_{ii}" in dataset.variables:
            counts_array = counts_array + dataset[f"{identifier}_counts_{ii}"]
        counts_array = counts_array.sortby(counts_array[f"{identifier}_state_{ii}"])
        ds_temp.update({f"{identifier}_counts_{ii}": counts_array})
        datasets.append(ds_temp)
    dataset_new = merge_datasets_dac(datasets)
    dataset_merged = dataset.merge(dataset_new, compat="override")
    return dataset_merged


def show_figure(fig):
    """
    Shows a closed figure.

    Args:
        fig: Figure to show.
    """
    dummy = plt.figure()
    new_manager = dummy.canvas.manager
    new_manager.canvas.figure = fig
    fig.set_canvas(new_manager.canvas)


class Benchmark(ABC):
    """
    A base class for running cortex-based Benchmark experiments.

    In order to write a new benchmark, it is recommended to derive a class
    from this baseclass. A new benchmark is defined by deriving from this base
    class and implementing the execute method. Additionally, a custom analysis
    for the benchmark can be implemented by giving the pointer to the analysis
    method in ``analysis_function`` field. The given analysis_function should
    accept ``AnalysisResult`` as its input and return the final result.
    """

    analysis_function = staticmethod(default_analysis_function)
    default_options: dict[str, Any] | None = None
    options: dict[str, Any] | None = None
    # name: str = "unnamed_benchmark"

    def __init__(self, backend: Union[str, IQMBackendBase], configuration: "BenchmarkConfigurationBase", **kwargs):

        # Ported from BenchmarkBase   # CHECK
        self.configuration = configuration
        self.serializable_configuration = deepcopy(self.configuration)
        self.serializable_configuration.benchmark = self.name

        if isinstance(backend, str):
            self.backend = get_iqm_backend(backend)
        else:
            assert isinstance(backend, IQMBackendBase)
            self.backend = backend

        self.shots = self.configuration.shots
        self.calset_id = self.configuration.calset_id
        self.max_gates_per_batch = self.configuration.max_gates_per_batch

        self.routing_method = self.configuration.routing_method
        self.physical_layout = self.configuration.physical_layout

        self.untranspiled_circuits: Dict[str, Dict[int, list]] = {}
        self.transpiled_circuits: Dict[str, Dict[int, list]] = {}

        # From exa_support MR
        self.options = copy.copy(self.default_options) if self.default_options else {}
        self.options.update(kwargs)
        self.runs: list[BenchmarkRunResult] = []

    @classmethod
    @abstractmethod
    def name(cls):
        """Return the name of the benchmark."""

    @abstractmethod
    def execute(self, backend: IQMBackend | IQMFacadeBackend | str) -> xr.Dataset:
        """
        Executes the benchmark.

        This method should be overridden by deriving classes.

        Args:
            backend: Qiskit backend used to execute benchmarks.

        Returns:
            An xarray dataset which contains to results of the benchmark execution.
            The dataset should contain all the information necessary for analysing
            the benchmark results.
        """

    def run(self, calibration_set_id: str | uuid.UUID | None = None) -> BenchmarkRunResult:
        """
        Runs the benchmark using the given backend and calibration_set_id.

        Args:
            calibration_set_id:
                CalibrationSetId used to initialize the backend or None for the latest calibration set.

        Returns:
            RunResult: The result of the benchmark run.
        """
        backend_for_execute = copy.copy(self.backend)
        backend_for_execute.run = functools.partial(
            self.backend.run, calibration_set_id=calibration_set_id
        )  # type: ignore
        dataset = self.execute(backend_for_execute)
        run = BenchmarkRunResult(dataset)
        self.runs.append(run)
        return run

    def analyze(self, run_index=-1) -> BenchmarkAnalysisResult:
        """
        The default analysis for the benchmark.

        Internally uses the function defined by the attribute ``analysis_function``
        to perform the analysis. This function makes a shallow copy of the dataset produced by
        run. Therefore, it is recommended to not make changes to the data of the dataset but
        just the structure of the array.

        Args:
            run_index: Index for the run to analyze.

        Returns:
            An analysis result constructed from the run and updated by the analysis method defined by
            the ``analysis_function`` field.
        """
        run = self.runs[run_index]
        result = BenchmarkAnalysisResult.from_run_result(run)
        updated_result = self.analysis_function(result)
        return updated_result



"""
Utility functions used by mGST modules
"""

import os
import random
import warnings

import numpy as np
import numpy.linalg as la
from scipy.linalg import expm, qr

from mGST.low_level_jit import Mp_norm_lower, MVE_lower, contract, local_basis


def transp(dim1, dim2):
    """Superoperator of a map that performs the transpose operation

    Parameters
    ----------
    dim1 : int
        First dimension of the matrix to be transposed
    dim2 : int
        Second dimension of the matrix to be transposed

    Returns
    -------
    T: numpy array
        Transpose superoperator that turns a dim1xdim2 matrix into a dim2xdim1 matrix

    Notes:
        The projection is done with respect to the canonical metrik.
    """
    id1 = np.eye(dim1)
    id2 = np.eye(dim2)
    T = np.einsum("il,jk->ijkl", id2, id1).reshape(dim1 * dim2, dim1 * dim2)
    return T


def randpsd(n, normalized="True"):
    """Generate a random positive semidefinite square matrix

    Parameters
    ----------
    n : int
        Number of matrix entries
    normalized : {Ture, False}, optional
        Controls if the output is trace normalized, defaults to "True"

    Returns
    -------
    mat: numpy array
        Random positive definite square matrix

    Notes:
        Eigenvalues are drawn uniformly from the interval [0,1);
        The basis is generated by diagonalizing a random hermitian matrix (see randHerm function).
    """
    dim = int(np.sqrt(n))
    H = randHerm(dim)
    U, _, _ = np.linalg.svd(H)
    evals = np.random.random_sample(dim)
    if normalized == "True":
        evals /= np.sum(evals)
    mat = np.dot(np.dot(U, np.diag(evals)), U.T.conj())
    mat = mat.reshape(-1)
    return mat


def randvec(n):
    """Generate vector with real and imaginary part drawn from the normal distribution

    Parameters
    ----------
    n : int
        Number of elements for the random vector

    Returns
    -------
    g: 1D numpy array
        Length n vector with complex entries whose real and imaginary part is independently drawn
        from the normal distribution with mean 0 and variance 1.
    """
    # randn(n) produces a random vector of length n with mean 0 and variance 1
    g = np.random.randn(n) + 1j * np.random.randn(n)
    g = g / np.linalg.norm(g)
    return g


def randHerm(n):
    """Generate random square hermitian matrix

    Parameters
    ----------
    n : int
        Matrix dimension

    Returns
    -------
    G: 2D numpy array
        Random hermitian matrix normalized in spectral norm

    Notes:
        First a matrix with random complex entries is generated (see randvec function).
        This matrix is then projected onto the space of hermitian
        matrices and normalized in spectral norm.
    """
    G = randvec(n * n).reshape(n, n)
    G = (G + G.T.conj()) / 2
    # ord=2 gives the spectral norm
    G = G / np.linalg.norm(G, ord=2)
    return G


def randHermGS(d, r):
    """Generates random set of operators that are hermiticity preserving

    Parameters
    ----------
    d : int
        Number of gates
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension

    Returns
    -------
    X: 3D numpy array
        Array where random hermiticity preserving operators are stacked along the first axis.

    Notes:
        The function randHerm is used to generate random hermitian Choi matrices,
        whose indies are then rearanged to obtain random hermiticity preserving superoperators.
    """
    dim = int(np.sqrt(r))
    X = np.zeros((r, d, r), dtype="complex")
    for i in range(d):
        H = randHerm(r).reshape(dim, dim, dim, dim)
        H = np.einsum("ijkl->jlik", H)
        X[:, i, :] = H.reshape(r, r)
    return X


def randU(n, a=1):
    """Generates random unitary from a random hermitian generator

    Parameters
    ----------
    n : int
        Matrix dimension of the unitary
    a : float
        Parameter to control the norm of the hermitian generator

    Returns
    -------
    U: 2D numpy array
        Matrix exponential of random hermitian matrix times the imaginary unit.
    """
    return expm(1j * a * randHerm(n)).astype(np.complex128)


def randU_Haar(n):
    """Return a Haar distributed random unitary

    Parameters
    ----------
    n : int
        Matrix dimension of the unitary

    Returns
    -------
    U: 2D numpy array
        Random unitary matrix distributed according to the Haar measure.
    """
    Z = np.random.randn(n, n) + 1.0j * np.random.randn(n, n)
    [Q, R] = qr(Z)
    D = np.diag(np.diagonal(R) / np.abs(np.diagonal(R)))
    return np.dot(Q, D)


def randKrausSet(d, r, rK, a=1):
    """Generates random set of Kraus operators

    Parameters
    ----------
    d : int
        Number of gates
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Number of Kraus operators per gate ("Kraus rank")
    a : float
        Parameter to control the norm of the hermitian generator and thereby
        how far the gates are from the identity

    Returns
    -------
    K: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.

    Notes:
        Let pdim be the physical dimension. Then a set of Kraus operators is generated
        by taking the first pdim columns of a random unitary of size rK*pdim.
        The random unitary is generated from a random hermitian matrix.
    """
    pdim = int(np.sqrt(r))
    K = np.zeros((d, rK, pdim, pdim)).astype(np.complex128)
    for i in range(d):
        K[i, :, :, :] += randU(pdim * rK, a)[:, :pdim].reshape(rK, pdim, pdim)
    return K


def randKrausSet_Haar(d, r, rK):
    """Generates random set of Kraus operators

    Parameters
    ----------
    d : int
        Number of gates
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Number of Kraus operators per gate ("Kraus rank")

    Returns
    -------
    K: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.

    Notes:
        Let pdim be the physical dimension. Then a set of Kraus operators is
        generated by taking the first pdim columns of a random unitary of size
        rK*pdim. The random unitary is generated according the the Haar measure.
    """
    pdim = int(np.sqrt(r))
    K = np.zeros((d, rK, pdim, pdim)).astype(np.complex128)
    for i in range(d):
        K[i, :, :, :] += randU_Haar(pdim * rK)[:, :pdim].reshape(rK, pdim, pdim)
    return K


def random_gs(d, r, rK, n_povm):
    """Generates a random gate using the Gaussian unitary ensemble, initial state and POVM

    Parameters
    ----------
    d : int
        Number of gates
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Number of Kraus operators per gate ("Kraus rank")
    n_povm : int
        Number of POVM-Elements

    Returns
    -------
    K: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    X: 3D numpy array
        Array where random CPT superoperators are stacked along the first axis.

    Notes:
        The Kraus operators are generated from random unitaries, see function randKrausSet
    """
    K = randKrausSet(d, r, rK).copy()
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    rho = randpsd(r).copy()
    A = randKrausSet(1, r, n_povm)[0].conj()
    E = np.array([(A[i].T.conj() @ A[i]).reshape(-1) for i in range(n_povm)]).copy()
    return K, X, E, rho


def random_gs_Haar(d, r, rK, n_povm):
    """Generates a random gate set with gates from Haar random unitaries, initial state and POVM

    Parameters
    ----------
    d : int
        Number of gates
    r : int
        Superoperator dimension of the gates given by the square of the physical dimension
    rK : int
        Number of Kraus operators per gate ("Kraus rank")
    n_povm : int
        Number of POVM-Elements

    Returns
    -------
    K: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    X: 3D numpy array
        Array where random CPT superoperators are stacked along the first axis.

    Notes:
        The Kraus operators are generated from Haar random unitaries, see function randKrausSet_Haar
    """
    K = randKrausSet_Haar(d, r, rK).copy()
    X = np.einsum("ijkl,ijnm -> iknlm", K, K.conj()).reshape((d, r, r))
    rho = randpsd(r).copy()
    A = randKrausSet_Haar(1, r, n_povm)[0].conj()
    E = np.array([(A[i].T.conj() @ A[i]).reshape(-1) for i in range(n_povm)]).copy()
    return K, X, E, rho


def perturbed_target_init(X_target, rK):
    """Generates a small random noise gate around the identity and applies it to the target gate
    The reason for using this gate as an initialization as opposed the the target gate itself, is
    that the non-dominant Kraus operators we start with are now not zero, but small random matrices.
    Observations show that with this start, the non-dominant Kraus operators converge faster.

    Parameters
    ----------
    X_target : 3D numpy array
        Current gate estimate
    rK : int
        Number of Kraus operators per gate ("Kraus rank") for the initialization

    Returns
    -------
    K_init: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.
    """
    d, r, _ = X_target.shape
    pdim = int(np.sqrt(r))
    K_perturb = randKrausSet(d, r, rK, a=0.1)
    X_perturb = np.einsum("ijkl,ijnm -> iknlm", K_perturb, K_perturb.conj()).reshape((d, r, r))
    X_init = np.einsum("ikl,ilm ->ikm", X_perturb, X_target)
    K_init = Kraus_rep(X_init, d, pdim, rK)
    return K_init


def basis(size, index):
    """Creates standard basis vectors

    Parameters
    ----------
    size : int
        Vector space dimension
    index : int
        Index of basis vector

    Returns
    -------
    vec: 1D numpy array
        Vector with entry 1 at position given by index and zeros elsewhere
    """
    vec = np.zeros(size)
    vec[index] = 1.0
    return vec


def depol(pdim, p):
    """Kraus representation of depolarizing channel

    Parameters
    ----------
    pdim : int
        Physical dimension
    p : float
        Error probability

    Returns
    -------
    K_depol: 4D numpy array
        Each entry along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.

    Notes:
        The depolarizing channel is defined as L(rho) = (1-p)*rho + p/pdim*Id.
    """
    phi_plus = np.sum([np.kron(basis(pdim, i), basis(pdim, i)) for i in range(pdim)], axis=0)
    choi_state = p / pdim * np.eye(pdim**2) + (1 - p) * np.kron(phi_plus, phi_plus.reshape(pdim**2, 1))
    K_depol = la.cholesky(choi_state)
    return K_depol.reshape(pdim, pdim, pdim**2).swapaxes(0, 2)


def varassign(v, X, E, rho, argument):
    """Assigns input to specified gate set variables

    Parameters
    ----------
    v : numpy array
        New set of variables
    X : numpy array
        Current gate estimate
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    argument : {"X", "E", "rho"}
        Which part of the gate set is updated

    Returns
    -------
    [.,.,.]: 3 element list
        List in the order [X,E,rho] where either X, E or rho is repaced by v,
        depending on the input to the "arguement" variable
    """
    if argument in ["X", "K"]:
        return [v, E, rho]
    if argument == "E":
        return [X, v, rho]

    return [X, E, v]


def batch(y, J, bsize):
    """Returns random batch of sequences and corresponding measurements

    Parameters
    ----------
    y : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    bsize : int
        Size of the batch (number of sequences)

    Returns
    -------
    y_b : numpy array
        Randomly subsampled columns of y
    J_b : numpy array
        Randomly subsampled rows of J in accordance the the columns selected in y_b
    """
    if y.shape[1] <= bsize:
        return y, J
    if bsize < 1:  # if batch size is given as ratio
        bsize = int(bsize * len(J) // 1)
    batchmask = np.array([1] * bsize + [0] * (len(J) - bsize))
    np.random.shuffle(batchmask)
    J_b = J[batchmask == 1]
    y_b = y[:, batchmask == 1]
    return y_b, J_b


def F_avg_X(X, K):
    """Returns the average gate fidelity between two gates given by a
    superoperator and a set of Kraus operators

    Parameters
    ----------
    X : 2D numpy array
        CPT superoperator of size (pysical dimension**2) x (pysical dimension **2)
    K : 3D numpy array
        Array of Kraus operators with size (Kraus rank) x (pysical dimension) x (pysical dimension)

    Returns
    -------
    Fid : float
        Average gate fidelity

    Notes:
    Not gauge optimization involved; Average gate fidelity is gauge dependent.
    """
    pdim = K.shape[2]
    d = K.shape[0]
    Fid_list = []
    for k in range(d):
        choi_inner_prod = np.einsum("imjl,pml,pij", X[k].reshape(pdim, pdim, pdim, pdim), K[k], K[k].conj())

        unitality_term = np.einsum("imkk,pml,pil", X[k].reshape(pdim, pdim, pdim, pdim), K[k], K[k].conj())
        Fid = (choi_inner_prod + unitality_term) / pdim / (pdim + 1)
        Fid_list.append(Fid)
    return np.average(np.real(Fid_list)), np.real(Fid_list)


def MVE(X_true, E_true, rho_true, X, E, rho, d, length, n_povm, samples=10000):
    """Mean varation error between the outputs of two gate sets on random sequences

    Parameters
    ----------
    X_true : numpy array
        Target gates
    E_true : numpy array
        Target POVM
    rho_true : numpy array
        Target initial state
    X : numpy array
        Current gate estimate
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    d : int
        Number of different gates in the gate set
    length : int
        Length of the test sequences
    n_povm : int
        Number of POVM elements
    samples : int
        Number of random gate sequences over which the mean variation error is computed

    Returns
    -------
    MVE : float
        Mean varaition error

    Notes:
        Sequences are drawn without replacement from initally d**l possibilities.
        For each sequence the total variation error of the two probability distribution
        over the POVM elements is computed. Afterwards the meean over these total
        variation errors is returned.
    """
    if samples == "all" or np.log(samples) / np.log(d) > length:
        J = np.random.randint(0, d, length * d**length).reshape(d**length, length)
    else:
        J = np.random.randint(0, d, length * samples).reshape(samples, length)
    return MVE_lower(X_true, E_true, rho_true, X, E, rho, J, n_povm)


def Mp_norm(X_true, E_true, rho_true, X, E, rho, d, length, n_povm, p, samples=10000):
    """Mean of the p-norm deviation between the outputs of two gate sets on random sequences

    Parameters
    ----------
    X_true : numpy array
        Target gates
    E_true : numpy array
        Target POVM
    rho_true : numpy array
        Target initial state
    X : numpy array
        Current gate estimate
    E : numpy array
        Current POVM estimate
    rho : numpy array
        Current initial state estimate
    d : int
        Number of different gates in the gate set
    length : int
        Length of the test sequences
    n_povm : int
        Number of POVM elements
    p : int
        Defines the l_p - norm that is used to compare probability distributions
    samples : int
        Number of random gate sequences over which the mean variation error is computed

    Returns
    -------
    MPE : float
        Mean l_p - norm error

    Notes:
        Sequences are drawn without replacement from initally d**l possibilities.
        For each sequence the l_p - norm error of the two probability distribution
        over the POVM elements is computed. Afterwards the meean over these total
        variation errors is returned.

    """
    if samples == "all" or np.log(samples) / np.log(d) > length:
        J = np.random.randint(0, d, length * d**length).reshape(d**length, length)
    else:
        J = np.random.randint(0, d, length * samples).reshape(samples, length)
    return Mp_norm_lower(X_true, E_true, rho_true, X, E, rho, J, n_povm, p)


def Kraus_rep(X, d, pdim, rK):
    """Compute the Kraus representations for all gates in the gate set

    Parameters
    ----------
    X : numpy array
        Current gate estimate
    d : int
        Number of gates
    pdim : int
        Physical dimension
    rK : int
        Target Kraus rank

    Returns
    -------
    K: 4D numpy array
        Each subarray along the first axis contains a set of Kraus operators.
        The second axis enumerates Kraus operators for a gate specified by the first axis.

    Notes:
        The Kraus representation is obtained from a singular value decomposition
        of the Choi matrix. If parameter rK is smaller than the true rank of the
        Choi matrix, a rank rK approximation is used. Approximations are only CPT
        in the special case where the original gate was already of rank <= rK.
    """
    X_choi = X.reshape(d, pdim, pdim, pdim, pdim)
    X_choi = np.einsum("ijklm->iljmk", X_choi).reshape((d, pdim**2, pdim**2))
    K = np.zeros((d, rK, pdim, pdim)).astype(np.complex128)
    for i in range(d):
        w, v = la.eigh(X_choi[i])
        if np.min(w) < -1e-12:
            raise ValueError("Choi Matrix is not positive definite within tolerance 1e-12")
        K[i] = np.einsum("ijk->kji", (v[:, -rK:] @ np.diag(np.sqrt(np.abs(w[-rK:])))).reshape(pdim, pdim, rK))
        # Trace normalization of Choi Matrix
        K[i] = K[i] / np.sqrt(np.einsum("ijk,ijk", K[i], K[i].conj())) * np.sqrt(pdim)
    return np.array(K)


def sampled_measurements(y, n):
    """Compute finite sample estimates of input probabilities

    Parameters
    ----------
    y : numpy array
        2D array of measurement outcomes for different sequences;
        Each column contains the outcome probabilities for a fixed sequence
    n : Number of samples for each experiment

    Returns
    -------
    y_sampled : numpy array
        2D array of sampled measurement outcomes for different sequences;
        Each column contains the sampled outcome probabilities for a fixed sequence

    Notes:
        The entries of each column of y form a probability distribution.
        From this distribution n random samples are drawn which give estimates for
        the initial probabilities. This simulates finite sample size data.
    """
    n_povm, m = y.shape
    if any(y.reshape(-1) > 1) or any(y.reshape(-1) < 0):
        y_new = np.maximum(np.minimum(y, 1), 0)
        if np.sum(np.abs(y_new - y)) > 1e-6:
            warnings.warn(
                f"Warning: Probabilities capped to interval [0,1]",
                f"l1-difference to input:%f" % np.sum(np.abs(y_new - y)),
            )
        y = y_new
    rng = np.random.default_rng()
    y_sampled = np.zeros(y.shape)
    for i in range(m):
        y_sampled[:, i] = rng.multinomial(n, [y[o, i] for o in range(n_povm)]) / n
    return y_sampled


def random_len_seq(d, min_l, max_l, N):
    """Generate random gate sequence instructions which contain sequences of different lengths,
    the lengths are drawn uniformly at random from (min_l, ..., max_l)

    Parameters
    ----------
    d : int
        Number of gates
    min_l : int
        Minimum sequence length
    max_l : int
        Maximum sequence length
    N : int
        Number of random sequences

    Returns
    -------
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence

    """
    seq_lengths = np.random.randint(min_l, max_l + 1, N)
    J = []
    for length in seq_lengths:
        j_curr = np.random.randint(0, d, length)
        J.append(list(np.pad(j_curr, (0, max_l - length), "constant", constant_values=-1)))
    return np.array(J)


def generate_fids(d, length, m_f):
    """Generate random fiducial sequencecs

    Parameters
    ----------
    d : int
        Number of gates
    length : int
        Total sequence length
    m_f : int
        Number of random fiducial sequences

    Returns
    -------
    J_fid : numpy array
        Sequence list of only the fiducial sequences
    J_fid2 : numpy array
        Sequence list for all combinations of two concatenated fiducial sequences
    J_meas : numpy array
        Sequence list for all combinations of fiducials seuqneces with a gate
        in between: fiducial1 -- gate -- fiducial2
    """
    fid_length = (length - 1) // 2
    fid = random.sample(range(d**fid_length), m_f).copy()
    J_fid = [list(local_basis(ind, d, fid_length)) for ind in fid]
    J_fid2 = np.array([seqL + seqR for seqL in J_fid for seqR in J_fid])
    J_meas = np.zeros((d, m_f**2, length), dtype="int")
    for k in range(d):
        J_meas[k] = np.array([seqL + [k] + seqR for seqL in J_fid for seqR in J_fid])
    return np.array(J_fid), J_fid2, J_meas


def is_positive(X, E, rho):
    """Print the results for checks whether a gate set is physical.

    This includes all positivity and normalization constraints.

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    """
    d, r, _ = X.shape
    pdim = int(np.sqrt(r))
    n_povm = E.shape[0]

    X_choi = X.reshape(d, pdim, pdim, pdim, pdim)
    X_choi = np.einsum("ijklm->iljmk", X_choi).reshape((d, r, r))

    eigvals = np.array([la.eigvals(X_choi[i]) for i in range(d)])
    partial_traces = np.einsum("aiikl -> akl", X.reshape(d, pdim, pdim, pdim, pdim))
    povm_eigvals = np.array([la.eigvals(E[i].reshape(pdim, pdim)) for i in range(n_povm)])
    if np.any(np.imag(eigvals.reshape(-1) > 1e-10)):
        print("Gates are not all hermitian.")
    else:
        for i in range(d):
            print(f"Gate %i positive:" % i, np.all(eigvals[i, :] > -1e-10))
            print(f"Gate %i trace preserving:" % i, la.norm(partial_traces[i] - np.eye(pdim)) < 1e-10)
    print(f"Initial state positive:", np.all(la.eigvals(rho.reshape(pdim, pdim)) > -1e-10))
    print(f"Initial state normalization:", np.trace(rho.reshape(pdim, pdim)))
    print(
        "fPOVM valid:",
        np.all(
            [
                la.norm(np.sum(E, axis=0).reshape(pdim, pdim) - np.eye(pdim)) < 1e-10,
                np.all(povm_eigvals.reshape(-1) > -1e-10),
            ]
        ),
    )


def tvd(X, E, rho, J, y_data):
    """Return the total variation distance between model probabilities for the circuits in J
    and the probabilities given by y_data.

    Parameters
    ----------
    X : numpy array
        Gate set
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    y_data : numpy array
        2D array of measurement outcomes for sequences in J;
        Each column contains the outcome probabilities for a fixed sequence
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence
    bsize : int
        Size of the batch (number of sequences)

    Returns
    -------
    dist : float
        The total variation distance.
    """
    n_povm = y_data.shape[0]
    y_model = np.real(np.array([[E[i].conj() @ contract(X, j) @ rho for j in J] for i in range(n_povm)]))
    dist = la.norm(y_model - y_data, ord=1) / 2
    return dist


def random_seq_design(
    d, l_min, l_cut, l_max, N_short, N_long
):  # Draws without replacement but ineficiently (not working for sequence length > 24)
    """Generate a set of random sequences ith given lengths
    This sequence lengths (circuit depths) are chosen for a mix of very short sequences (better convergence) and some
    slightly longer sequences to reduce the generalization error.
    This sequence design is heuristic and intended for coarse and fast estimates. For very accurate estimates at the
    cost of higher measurement effort it is recommended to use pyGSTi with long sequence GST.

    Parameters
    ----------
    d : int
        The number of gates in the gate set
    l_min : int
        Minimum sequence lenght
    l_cut : int
        Cutoff sequence lenght: N_short sequences are equally distributed among lengths l_min < l < l_cut
    l_max : int
        N_long sequences are equally distributed among lengths l_cut + 1 < l < l_max. Currently l_max < 24 only.
    N_short : int
        Number of short sequences
    N_long : int
        Number of long sequences

    Returns
    -------
    J : numpy array
        2D array where each row contains the gate indices of a gate sequence

    """
    # Open problems:
    # - Change randomness to work with longer sequences;
    # - Handle case where Number of sequences is smaller than the available lengths
    if l_max >= 24:
        raise ValueError("Currently only sequences lenghts < 24 are supported.")

    J = np.array([-np.ones(l_max)])
    # Short sequences:
    seq_counts = []
    N_remaining = N_short
    for l in range(l_min, l_cut + 1):
        seq_counts.append(int(np.min([np.floor(N_remaining / (l_cut + 1 - l)), d**l])))
        if seq_counts[-1] > 0:
            ind_curr = np.array(random.sample(range(d**l), seq_counts[-1]))
            J_curr = np.array(
                [np.pad(local_basis(ind, d, l), (0, l_max - l), "constant", constant_values=-1) for ind in ind_curr]
            )
            J = np.concatenate((J, J_curr), axis=0)
            N_remaining = N_short - 1 - np.sum(seq_counts)
    if N_remaining > 0:
        print(
            "Number of possible sequences without replacement for the given sequence\
        length range is lower than the desired total number of sequences"
        )

    # Long sequences:
    seq_counts = []
    N_remaining = N_long
    for l in range(l_cut + 1, l_max + 1):
        seq_counts.append(int(np.min([np.floor(N_remaining / (l_max + 1 - l)), d**l])))
        if seq_counts[-1] > 0:
            ind_curr = np.array(random.sample(range(d**l), seq_counts[-1]))
            J_curr = np.array(
                [np.pad(local_basis(ind, d, l), (0, l_max - l), "constant", constant_values=-1) for ind in ind_curr]
            )
            J = np.concatenate((J, J_curr), axis=0)
            N_remaining = N_long - np.sum(seq_counts)
    return J.astype(int)


def n_params(pdim, d, rK, n_povm):
    """Returns the number of free parameters in a gate set
    Unitary gauge parameters and unitary freedom in representing the Kraus operators and the POVM elements are
    subtracted. For details, see
    https://docserv.uni-duesseldorf.de/servlets/DerivateServlet/Derivate-71735/dissertation_brieger-2%281%29.pdf

    Parameters
    ----------
    pdim : int
        The physical dimension, i.e. 2^(#qubits)
    d : int
        The number of gates in the gate set
    rK : int
        The Kraus rank of the gates
    n_povm : int
        The number of POVM elements; For computational basis measurements n_povm = pdim

    Returns
    -------
    n_params : int
        The number of free parameters
    """
    # Order: gates + stat + povm - povm_freedom - gauge freedom - Kraus freedom
    return d * (pdim**2 * (2 * rK - 1) - rK**2) + pdim * (n_povm * pdim - pdim)


def multikron(matrix_array):
    """Computes the Kronecker product of all matrices in an array
    The order is matrix_array[0] otimes matrix_array[1] otimes ...

    Parameters
    ----------
    matrix_array: numpy array
        An array containing matrices

    Returns
    -------
    res: numpy array
        The resulting tensor product (potentially a very large array)
    """
    res = matrix_array[0]
    for i in range(1, matrix_array.shape[0]):
        res = np.kron(res, matrix_array[i])
    return res


def local_dephasing_pp(prob_vec):
    """Returns the tensor product of single qubit dephasing channels in Pauli basis

    Parameters
    ----------
    prob_vec : list[float]
        A list of dephasing probabilities

    Returns
    -------
    D_final : numpy array
        Process matrix of the tensor product of local dephasing channels

    """
    D_loc_array = np.array([[[1, 0, 0, 0], [0, 1 - 2 * p, 0, 0], [0, 0, 1 - 2 * p, 0], [0, 0, 0, 1]] for p in prob_vec])
    D_final = multikron(D_loc_array)
    return D_final


def outcome_probs_from_files(folder_name, basis_dict, n_povm, N):
    """Searches a specified folder for .txt files containing circuit outcomes and combines the results
    Each text file needs to have line in the following format:
    1: 1,0,1,0,1,1,0,0,0
    2: 1,1,1,1,0,1,1,1,1
    Here the first number specifies the circuit and basis outcome of each shot is written to the right.

    Parameters
    ----------
    folder_name : str
        The relative or absolute name of the data folder
    basis_dict : dict[str: int]
        Translation between label for each shot in the .txt files and the numbering of the POVM element.
        Example (two qubits): {"00": 0, "01": 1, "10": 2, "11": 3,}
    n_povm : int
        The number of POVM elements in the data set
    N : int
        The number of circuits in the data set

    Returns
    -------
    y : numpy array
        2D array of measurement outcomes for all measured sequences;
        Each column contains the outcome probabilities for a fixed sequence
    avg_counts : int
        The average number of shots per circuit after combining data sets
    """
    filenames = os.listdir(path=folder_name)
    datafile_names = [s for s in filenames if ".txt" in s]

    # array of outcome probabilities:
    y = np.zeros((n_povm, N))
    sample_counts = np.zeros((len(datafile_names), N))
    k = 0
    for filename in datafile_names:
        with open(folder_name + "/" + filename, encoding="ASCII") as file:
            i = 0
            for line in file:
                # removing row index at beginning and \n mark at end
                line_entries = line.rstrip().split(": ")
                # splitting result string at commas
                result_list = line_entries[1:][0].split(",")
                # translate each measurement result onto basis index
                for entry in result_list:
                    j = basis_dict[entry]
                    y[j, i] += 1
                sample_counts[k, i] = len(result_list)
                i += 1
        k += 1
    total_counts = np.sum(sample_counts, axis=0)
    avg_counts = int(np.average(total_counts))
    return y / total_counts, avg_counts



"""Tests for mirror RB"""

import numpy as np

from iqm.benchmarks.randomized_benchmarking.clifford_rb.clifford_rb import (
    CliffordRandomizedBenchmarking,
    CliffordRBConfiguration,
)
from iqm.benchmarks.randomized_benchmarking.interleaved_rb.interleaved_rb import (
    InterleavedRandomizedBenchmarking,
    InterleavedRBConfiguration,
)
from iqm.benchmarks.randomized_benchmarking.mirror_rb.mirror_rb import (
    MirrorRandomizedBenchmarking,
    MirrorRBConfiguration,
)


backend = "fakeapollo"


class TestRB:
    def test_mrb(self):
        EXAMPLE_MRB = MirrorRBConfiguration(
            qubits_array=[[0, 1], [0, 3]],
            depths_array=[[2**m for m in range(4)]],
            num_circuit_samples=2,
            num_pauli_samples=2,
            shots=2**4,
            qiskit_optim_level=1,
            routing_method="sabre",
            two_qubit_gate_ensemble={"CZGate": 0.8, "iSwapGate": 0.2},
            density_2q_gates=0.25,
        )
        benchmark = MirrorRandomizedBenchmarking(backend, EXAMPLE_MRB)
        benchmark.run()
        benchmark.analyze()

    def test_irb(self):
        EXAMPLE_IRB_1Q = InterleavedRBConfiguration(
            qubits_array=[[0]],
            sequence_lengths=[2 ** (m + 1) - 1 for m in range(4)],
            num_circuit_samples=2,
            shots=2**4,
            parallel_execution=True,
            interleaved_gate="RGate",
            interleaved_gate_params=[np.pi, 0],
            simultaneous_fit=["amplitude", "offset"],
        )
        benchmark = InterleavedRandomizedBenchmarking(backend, EXAMPLE_IRB_1Q)
        benchmark.run()
        benchmark.analyze()

    def test_crb(self):
        EXAMPLE_CRB_1Q = CliffordRBConfiguration(
            qubits_array=[[3]],
            sequence_lengths=[2 ** (m + 1) - 1 for m in range(4)],
            num_circuit_samples=2,
            shots=2**4,
            calset_id=None,
            parallel_execution=False,
        )
        benchmark = CliffordRandomizedBenchmarking(backend, EXAMPLE_CRB_1Q)
        benchmark.run()
        benchmark.analyze()



"""
Interface to pyGSTi
"""

import numpy as np
from pygsti.baseobjs import Basis, Label
from pygsti.circuits import create_lsgst_circuits
from pygsti.models.modelconstruction import create_explicit_model_from_expressions
from pygsti.report.reportables import entanglement_fidelity
from pygsti.tools import change_basis


def pygsti_model_to_arrays(model, basis="pp"):
    """Turn the gate set of a pygsti model into numpy arrays used by mGST.

    Parameters
    ----------
    model : pygsti ExplicitOpModel object
        Contains the model parameters in the Pauli transfer matrix formalism

    basis : {'pp','std'}
        The basis in which the output gate set will be given. It can be either
        'pp' (Pauli basis) or 'std' (standard basis)

    Returns
    -------
    X : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
    E : numpy array
        POVM matrix of shape (#POVM elements, dimension^2)
    rho : numpy array
        Initial state vector of shape (dimension^2)
    """
    X = []
    op_Labels = list(model.__dict__["operations"].keys())
    effect_Labels = list(model["Mdefault"].keys())
    E = np.array([model["Mdefault"][label].to_dense().reshape(-1) for label in effect_Labels])
    rho = model["rho0"].to_dense().reshape(-1)
    for op_Label in op_Labels:
        X.append(model[op_Label].to_dense())
    if basis == "pp":
        return (np.array(X).astype(np.complex128), E.astype(np.complex128), rho.astype(np.complex128))
    if basis == "std":
        return pp2std(np.array(X), E, rho)
    raise ValueError(f"Wrong basis input, use either std or pp")


def average_gate_fidelities(model1, model2, pdim, basis_string="pp"):
    """Return the average gate fidelities between gates of two pygsti models.

    Parameters
    ----------
    model : pygsti ExplicitOpModel object
        Contains the model parameters in the Pauli transfer matrix formalism
    model2 : pygsti ExplicitOpModel object
        Contains the model parameters in the Pauli transfer matrix formalism
    pdim : int
        physical dimension
    basis : {'pp','std'}
        The basis in which the input models are gíven. It can be either
        'pp' (Pauli basis) or 'std' (standard basis, Default)

    Returns
    -------
    fidelities : 1D numpy array
        Array containing the average gate fidelities for all gates
    """
    ent_fids = []
    basis = Basis.cast(basis_string, pdim**2)
    labels1 = list(model1.__dict__["operations"].keys())
    labels2 = list(model2.__dict__["operations"].keys())

    for i, label in enumerate(labels1):
        ent_fids.append(float(entanglement_fidelity(model1[label], model2[labels2[i]], basis)))
    fidelities = (np.array(ent_fids) * pdim + 1) / (pdim + 1)
    return fidelities


def model_agfs(model, pdim):
    """Return the average gate fidelities between gates of two pygsti models.

    Parameters
    ----------
    model : pygsti ExplicitOpModel object
        Contains the model parameters in the Pauli transfer matrix formalism
    pdim : int
        physical dimension

    Returns
    -------
    fidelities : 1D numpy array
        Average gate fidelities between all different gates of the input model
    """
    ent_fids = []
    basis = Basis.cast("pp", pdim**2)
    labels = list(model.__dict__["operations"].keys())
    for i, _ in enumerate(labels):
        for j, _ in enumerate(labels):
            if j > i:
                ent_fids.append(float(entanglement_fidelity(model[labels[i]], model[labels[j]], basis)))
    fidelities = (np.array(ent_fids) * pdim + 1) / (pdim + 1)
    return fidelities


def arrays_to_pygsti_model(X, E, rho, basis="std"):
    """Turns a gate set given by numpy arrays into a pygsti model

    pygsti model is by default in Pauli-basis

    Parameters
    ----------
    X : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
    E : numpy array
        POVM matrix of shape (#POVM elements, dimension^2)
    rho : numpy array
        Initial state vector of shape (dimension^2)
    basis : {'pp','std'}
        The basis in which the INPUT gate set is given. It can be either
        'pp' (Pauli basis) or 'std' (standard basis)

    Returns
    -------
    model : pygsti ExplicitOpModel object
        Contains the model parameters in the Pauli transfer matrix formalism
    """
    d = X.shape[0]
    pdim = int(np.sqrt(len(rho)))
    Id = np.zeros(pdim**2)
    Id[0] = 1
    effect_label_str = [f"%i" % k for k in range(E.shape[0])]
    if basis == "std":
        X, E, rho = std2pp(X, E, rho)
    mdl_out = create_explicit_model_from_expressions(
        list(range(int(np.log(pdim) / np.log(2)))),
        [Label(f"G%i" % i) for i in range(d)],
        [f":".join([f"I(%i)" % i for i in range(int(np.log(pdim) / np.log(2)))]) for length in range(d)],
        effect_labels=effect_label_str,
    )
    mdl_out["rho0"] = np.real(rho)
    mdl_out["Mdefault"].from_vector(np.real(E).reshape(-1))
    for i in range(d):
        mdl_out[Label(f"G%i" % i)] = np.real(X[i])
    return mdl_out


def std2pp(X, E, rho):
    """Basis change of an mGST model from the standard basis to the Pauli basis

    Parameters
    ----------
    X : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
        in standard basis
    E : numpy array
        POVM matrix of shape (#POVM elements, dimension^2) in standard basis
    rho : numpy array
        Initial state vector of shape (dimension^2) in standard basis

    Returns
    -------
    Xpp : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
        in Pauli basis
    Epp : numpy array
        POVM matrix of shape (#POVM elements, dimension^2) in Pauli basis
    rhopp : numpy array
        Initial state vector of shape (dimension^2) in Pauli basis
    """
    Xpp = np.array([np.array(change_basis(X[i], "std", "pp")) for i in range(X.shape[0])])
    Epp = np.array([np.array(change_basis(E[i], "std", "pp")) for i in range(E.shape[0])])
    return (Xpp.astype(np.complex128), Epp.astype(np.complex128), change_basis(rho, "std", "pp").astype(np.complex128))


def pp2std(X, E, rho):
    """Basis change of an mGST model from the Pauli basis to the standard basis.

    Parameters
    ----------
    X : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
        in Pauli basis
    E : numpy array
        POVM matrix of shape (#POVM elements, dimension^2) in Pauli basis
    rho : numpy array
        Initial state vector of shape (dimension^2) in Pauli basis

    Returns
    -------
    Xstd : numpy array
        Gate set tensor of shape (Number of Gates, Kraus rank, dimension^2, dimension^2)
        in standard basis
    Estd : numpy array
        POVM matrix of shape (#POVM elements, dimension^2) in standard basis
    rhostd : numpy array
        Initial state vector of shape (dimension^2) in standard
    """
    Xstd = np.array([np.array(change_basis(X[i], "pp", "std")) for i in range(X.shape[0])])
    Estd = np.array([np.array(change_basis(E[i], "pp", "std")) for i in range(E.shape[0])])
    return (
        Xstd.astype(np.complex128),
        Estd.astype(np.complex128),
        change_basis(rho, "pp", "std").astype(np.complex128),
    )


def pygstiExp_to_list(model, max_germ_len):
    """Takes the sequences of a pyGSTi model and turns them into a sequence list for mGST.

    Parameters
    ----------
    model : pygsti ExplicitOpModel object
        The model containing the gate set parameters in the Pauli transfer matrix formalism.
    max_germ_len : int
        The maximum number of germ repetitions in the pyGSTi circuit design.

    Returns
    -------
    J_GST : numpy array, shape (num_experiments, max_sequence_length)
        A 2D array where each row contains the gate indices of a gate sequence. Each gate index
        corresponds to a specific gate in the model's target gate set.
    """
    prep_fiducials = model.prep_fiducials()
    meas_fiducials = model.meas_fiducials()
    germs = model.germs()
    maxLengths = [max_germ_len]
    listOfExperiments = create_lsgst_circuits(model.target_model(), prep_fiducials, meas_fiducials, germs, maxLengths)
    op_Labels = list(model.target_model().__dict__["operations"].keys())
    exp_list = []
    max_length = max((len(x.to_pythonstr(op_Labels)) for x in listOfExperiments))
    for i, exp_current in enumerate(listOfExperiments):
        exp = exp_current.to_pythonstr(op_Labels)
        gate_numbers = [ord(x) - 97 for x in exp.lower()]
        gate_numbers = np.pad(gate_numbers, (0, max_length - len(gate_numbers)), "constant", constant_values=-1)
        exp_list.append(gate_numbers)

    J_GST = np.array([[int(exp_list[i][j]) for j in range(max_length)] for i in range(len(exp_list))])
    return J_GST



"""
Generation of figures
"""

from matplotlib import ticker
from matplotlib.colors import Normalize
import matplotlib.pyplot as plt
import numpy as np
import numpy.linalg as la


SMALL_SIZE = 8
MEDIUM_SIZE = 9
BIGGER_SIZE = 10

plt.rc("font", size=SMALL_SIZE)  # controls default text sizes
plt.rc("axes", titlesize=SMALL_SIZE)  # fontsize of the axes title
plt.rc("axes", labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
plt.rc("xtick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("ytick", labelsize=SMALL_SIZE)  # fontsize of the tick labels
plt.rc("legend", fontsize=SMALL_SIZE)  # legend fontsize
plt.rc("figure", titlesize=BIGGER_SIZE)  # fontsize of the figure title

cmap = plt.colormaps.get_cmap("RdBu")
norm = Normalize(vmin=-1, vmax=1)


def set_size(w, h, ax=None):
    """Forcing a figure to a specified size

    Parameters
    ----------
    w : floar
        width in inches
    h : float
        height in inches
    ax : matplotlib axes
        The optional axes
    """
    if not ax:
        ax = plt.gca()
    l = ax.figure.subplotpars.left
    r = ax.figure.subplotpars.right
    t = ax.figure.subplotpars.top
    b = ax.figure.subplotpars.bottom
    figw = float(w) / (r - l)
    figh = float(h) / (t - b)
    ax.figure.set_size_inches(figw, figh)


def plot_objf(res_list, delta, title):
    """Plots the objective function over iterations in the algorithm

    Parameters:
    res_list (List[float]): The residual values
    delta (float): The success threshold
    title (str): The plot title

    Returns:
    """
    plt.semilogy(res_list)
    plt.ylabel(f"Objective function")
    plt.xlabel(f"Iterations")
    plt.axhline(delta, color="green", label="conv. threshold")
    plt.title(title)
    plt.legend()
    plt.show()


def generate_spam_err_pdf(filename, E, rho, E2, rho2, title=None, spam2_content="ideal"):
    """Generate pdf plots of two sets of POVM + state side by side in vector shape - Pauli basis
    The input sets can be either POVM/state directly or a difference different SPAM parametrizations to
    visualize errors.

    Parameters
    ----------
    filename : str
        The name under which the figures are saved in format "folder/name"
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    E2 : numpy array
        POVM #2
    rho2 : numpy array
        Initial state #2
    title : str
        The Figure title
    basis_labels : list[str]
        A list of labels for the basis elements. For the Pauli basis ["I", "X", "Y", "Z"] or the multi-qubit version.
    spam2_content : str
        Label of the right SPAM plot to indicate whether it is the ideal SPAM parametrization or for instance
        the error between the reconstructed and target SPAM

    Returns
    -------
    """
    r = rho.shape[0]
    pdim = int(np.sqrt(r))
    n_povm = E.shape[0]
    fig, axes = plt.subplots(ncols=2, nrows=n_povm + 1, sharex=True)
    plt.rc("image", cmap="RdBu")

    ax = axes[0, 0]
    ax.imshow(rho, vmin=-1, vmax=1)  # change_basis(S_true_maps[0],"std","pp")
    ax.set_xticks(np.arange(r))
    ax.set_title(r"rho")
    ax.yaxis.set_major_locator(ticker.NullLocator())

    ax = axes[0, 1]
    im0 = ax.imshow(rho2, vmin=-1, vmax=1)  # change_basis(S_true_maps[0],"std","pp")
    ax.set_xticks(np.arange(r))
    ax.set_title(r"rho - " + spam2_content)
    ax.yaxis.set_major_locator(ticker.NullLocator())

    for i in range(n_povm):
        ax = axes[1 + i, 0]
        ax.imshow(E[i], vmin=-1, vmax=1)  # change_basis(S_true_maps[0],"std","pp")
        ax.set_xticks(np.arange(pdim))
        ax.set_xticklabels(np.arange(pdim) + 1)
        ax.set_title(f"E%i" % (i + 1))
        ax.yaxis.set_major_locator(ticker.NullLocator())
        ax.xaxis.set_major_locator(ticker.NullLocator())

        ax = axes[1 + i, 1]
        ax.imshow(E2[i], vmin=-1, vmax=1)  # change_basis(S_true_maps[0],"std","pp")
        ax.set_xticks(np.arange(pdim))
        ax.set_xticklabels(np.arange(pdim) + 1)
        ax.set_title(f"E%i - " % (i + 1) + spam2_content)
        ax.yaxis.set_major_locator(ticker.NullLocator())
        ax.xaxis.set_major_locator(ticker.NullLocator())

    cbar = fig.colorbar(im0, ax=axes.ravel().tolist(), pad=0.1)
    cbar.ax.set_ylabel(r"Pauli basis coefficient", labelpad=5, rotation=90)

    if title:
        fig.suptitle(title)
    if r > 16:
        fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes, pad=0, shrink=0.6)
        fig.subplots_adjust(left=0, right=0.7, top=0.90, bottom=0.05, wspace=-0.6, hspace=0.4)
    else:
        fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes, pad=0)
        fig.subplots_adjust(left=0, right=0.7, top=0.90, bottom=0.05, wspace=-0.6, hspace=0.8)

    set_size(3, 2)
    plt.savefig(filename, dpi=150, transparent=True)
    plt.close()


def generate_spam_err_std_pdf(
    filename, E, rho, E2, rho2, basis_labels=False, title=None, magnification=10, return_fig=False
):
    """Generate pdf plots of two sets of POVM + state side by side in matrix shape - standard basis
    The input sets can be either POVM/state directly or a difference different SPAM parametrizations to
    visualize errors.

    Parameters
    ----------
    filename : str
        The name under which the figures are saved in format "folder/name"
    E : numpy array
        POVM
    rho : numpy array
        Initial state
    E2 : numpy array
        POVM #2
    rho2 : numpy array
        Initial state #2
    title : str
        The Figure title
    basis_labels : list[str]
        A list of labels for the basis elements. For the standard basis this could be ["00", "01",...]
    magnification : float
        A factor to be applied to magnify errors in the rightmost plot.
    return_fig : bool
        If set to True, a figure object is returned by the function, otherwise the plot is saved as <filename>
    Returns
    -------
    """
    dim = rho.shape[0]
    pdim = int(np.sqrt(dim))
    n_povm = E.shape[0]

    fig, axes = plt.subplots(ncols=3, nrows=n_povm + 1, gridspec_kw={"width_ratios": [1, 1, 1]}, sharex=True)
    plt.rc("image", cmap="RdBu")

    for i in range(n_povm + 1):
        if i == 0:
            plot_matrices = [np.real(rho), np.real(rho2), np.real(rho - rho2)]
            axes[i, 0].set_ylabel(f"rho", rotation=90, fontsize="large")
        else:
            plot_matrices = [np.real(E[i - 1]), np.real(E2[i - 1]), np.real(E[i - 1] - E2[i - 1]) * magnification]
            axes[i, 0].set_ylabel(f"E_%i" % (i - 1), rotation=90, fontsize="large")

        for j in range(3):
            ax = axes[i, j]
            ax.patch.set_facecolor("whitesmoke")
            ax.set_aspect("equal")
            for (x, y), w in np.ndenumerate(plot_matrices[j].reshape(pdim, pdim)):
                size = np.sqrt(np.abs(w))
                rect = plt.Rectangle(
                    [x + (1 - size) / 2, y + (1 - size) / 2],
                    size,
                    size,
                    facecolor=cmap((w + 1) / 2),
                    edgecolor=cmap((w + 1) / 2),
                )
                # print(cmap(size))
                ax.add_patch(rect)
            ax.invert_yaxis()
            ax.set_xticks(np.arange(pdim + 1), labels=[])

            ax.set_yticks(np.arange(pdim + 1), labels=[])
            ax.tick_params(which="major", length=0)  # Turn dummy ticks invisible
            ax.tick_params(which="minor", top=True, labeltop=True, bottom=False, labelbottom=False, length=0, pad=1)

            if pdim > 4:
                ax.grid(visible="True", alpha=0.4, lw=0.1)
                ax.set_xticks(np.arange(pdim) + 0.5, minor=True, labels=basis_labels, rotation=45, fontsize=2)
                ax.set_yticks(np.arange(pdim) + 0.5, minor=True, labels=basis_labels, fontsize=2)
            else:
                ax.grid(visible="True", alpha=0.4)
                ax.set_xticks(np.arange(pdim) + 0.5, minor=True, labels=basis_labels, rotation=45, fontsize=6)
                ax.set_yticks(np.arange(pdim) + 0.5, minor=True, labels=basis_labels, fontsize=6)
    if title:
        fig.suptitle(title)

    if dim > 16:
        fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes, pad=0, shrink=0.6)
        fig.subplots_adjust(left=0, right=0.7, top=0.90, bottom=0.05, wspace=-0.6, hspace=0.4)
    else:
        fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes, pad=0)
        fig.subplots_adjust(left=0, right=0.7, top=0.90, bottom=0.05, wspace=-0.6, hspace=0.8)

    if return_fig:
        return fig

    plt.savefig(filename, dpi=150, transparent=True)
    plt.close()
    return None


def generate_gate_err_pdf(
    filename, gates1, gates2, basis_labels=False, gate_labels=False, magnification=5, return_fig=False
):
    """Main routine to generate plots of reconstructed gates, ideal gates and the noise channels
    of the reconstructed gates.
    The basis is arbitrary but using gates in the Pauli basis is recommended.

    Parameters
    ----------
    filename : str
        The name under which the figures are saved in format "folder/name"
    gates1 : numpy array
        A gate set in the same format as the "X"-tensor. These gates are assumed to be the GST estimates.
    gates1 : numpy array
        A gate set in the same format as the "X"-tensor. These are assumed to be the target gates.
    basis_labels : list[str]
        A list of labels for the basis elements. For the standard basis this could be ["00", "01",...]
        and for the Pauli basis ["I", "X", "Y", "Z"] or the multi-qubit version.
    gate_labels : list[str]
        A list of names for the gates
    magnification : float
        A factor to be applied to magnify errors in the rightmost plot.
    return_fig : bool
        If set to True, a figure object is returned by the function, otherwise the plots are saved as <filename>
    """
    d = gates1.shape[0]
    dim = gates1[0].shape[0]
    if not basis_labels:
        basis_labels = np.arange(dim)
    if not gate_labels:
        gate_labels = [f"G%i" % k for k in range(d)]
    plot3_title = r"id - G U^{-1}"

    figures = []
    for i in range(d):
        if dim > 16:
            fig, axes = plt.subplots(ncols=1, nrows=3, gridspec_kw={"height_ratios": [1, 1, 1]}, sharex=True)
        else:
            fig, axes = plt.subplots(ncols=3, nrows=1, gridspec_kw={"width_ratios": [1, 1, 1]}, sharex=True)
        dim = gates1[0].shape[0]
        plot_matrices = [
            np.real(gates1[i]),
            np.real(gates2[i]),
            magnification * (np.eye(dim) - np.real(gates1[i] @ la.inv(gates2[i]))),
        ]

        for j in range(3):
            ax = axes[j]
            ax.patch.set_facecolor("whitesmoke")
            ax.set_aspect("equal")
            for (x, y), w in np.ndenumerate(plot_matrices[j].T):
                size = np.sqrt(np.abs(w))
                rect = plt.Rectangle(
                    [x + (1 - size) / 2, y + (1 - size) / 2],
                    size,
                    size,
                    facecolor=cmap((w + 1) / 2),
                    edgecolor=cmap((w + 1) / 2),
                )
                ax.add_patch(rect)
            ax.invert_yaxis()
            ax.set_xticks(np.arange(dim + 1), labels=[])
            ax.set_yticks(np.arange(dim + 1), labels=[])

            if dim > 16:
                ax.grid(visible="True", alpha=0.4, lw=0.1)
                ax.set_xticks(np.arange(dim) + 0.5, minor=True, labels=basis_labels, fontsize=2, rotation=45)
            else:
                ax.grid(visible="True", alpha=0.4)
                ax.set_xticks(np.arange(dim) + 0.5, minor=True, labels=basis_labels, rotation=45)
            ax.set_yticks(np.arange(dim) + 0.5, minor=True, labels=basis_labels)
            ax.tick_params(which="major", length=0)  # Turn dummy ticks invisible
            ax.tick_params(which="minor", top=True, labeltop=True, bottom=False, labelbottom=False, length=0)

        axes[0].set_title(r"G (estimate)", fontsize="large")
        axes[0].set_ylabel(gate_labels[i], rotation=90, fontsize="large")
        axes[1].set_title(r"U (ideal gate)", fontsize="large")
        axes[2].set_title(plot3_title, fontsize="large")
        fig.suptitle(f"Process matrices in the Pauli basis", va="bottom")

        if dim > 16:
            fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes.tolist(), pad=0, shrink=0.6)
            fig.subplots_adjust(left=0.1, right=0.76, top=0.85, bottom=0.03)
            set_size(0.5 * np.sqrt(dim), 1.3 * np.sqrt(dim))
        else:
            fig.colorbar(plt.cm.ScalarMappable(norm=norm, cmap=cmap), ax=axes.tolist(), pad=0)
            fig.subplots_adjust(left=0.1, right=0.76, top=0.85, bottom=0.03, hspace=0.2)
            set_size(2 * np.sqrt(dim), 0.8 * np.sqrt(dim))
        figures.append(fig)
        if not return_fig:
            plt.savefig(filename + f"G%i.pdf" % i, dpi=150, transparent=True, bbox_inches="tight")
    return figures



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Common functions for Randomized Benchmarking-based techniques
"""

from importlib import import_module
from itertools import chain
import os
import pickle
import random
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

from lmfit import Parameters, minimize
from lmfit.minimizer import MinimizerResult
from matplotlib.collections import PolyCollection
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
from qiskit import QuantumCircuit, transpile
from qiskit.quantum_info import Clifford
import xarray as xr

from iqm.benchmarks.logging_config import qcvv_logger
from iqm.benchmarks.randomized_benchmarking.multi_lmfit import create_multi_dataset_params, multi_dataset_residual
from iqm.benchmarks.utils import get_iqm_backend, marginal_distribution, submit_execute, timeit
from iqm.qiskit_iqm import optimize_single_qubit_gates
from iqm.qiskit_iqm.iqm_backend import IQMBackendBase


def compute_inverse_clifford(qc_inv: QuantumCircuit, clifford_dictionary: Dict) -> Optional[QuantumCircuit]:
    """Function to compute the inverse Clifford of a circuit
    Args:
        qc_inv (QuantumCircuit): The Clifford circuit to be inverted
        clifford_dictionary (Dict): A dictionary of Clifford gates labeled by (de)stabilizers
    Returns:
        Optional[QuantumCircuit]: A Clifford circuit
    """
    label_inv = str(Clifford(qc_inv).adjoint().to_labels(mode="B"))
    compiled_inverse = cast(dict, clifford_dictionary)[label_inv]
    return compiled_inverse


def estimate_survival_probabilities(num_qubits: int, counts: List[Dict[str, int]]) -> List[float]:
    """Compute a result's probability of being on the ground state.
    Args:
        num_qubits (int): the number of qubits
        counts (List[Dict[str, int]]): the result of the execution of a list of quantum circuits (counts)

    Returns:
        List[float]: the ground state probabilities of the RB sequence
    """
    return [c["0" * num_qubits] / sum(c.values()) if "0" * num_qubits in c.keys() else 0 for c in counts]


def exponential_rb(
    depths: np.ndarray, depolarization_probability: float, offset: float, amplitude: float
) -> np.ndarray:
    """Fit function for interleaved/non-interleaved RB

    Args:
        depths (np.ndarray): the depths of the RB experiment
        depolarization_probability (float): the depolarization value (1-p) of the RB decay
        offset (float): the offset of the RB decay
        amplitude (float): the amplitude of the RB decay

    Returns:
        np.ndarray: the exponential fit function
    """
    return (amplitude - offset) * (1 - depolarization_probability) ** depths + offset


def fit_decay_lmfit(
    func: Callable,
    qubit_set: List[int],
    data: List[List[float]] | List[List[List[float]]],
    rb_identifier: str,
    simultaneous_fit_vars: Optional[List[str]] = None,
) -> Tuple[np.ndarray, Parameters]:
    """Perform a fitting routine for 0th-order (Ap^m+B) RB using lmfit

    Args:
        func (Callable): the model function for fitting
        qubit_set (List[int]): the qubits entering the model
        data (List[List[float]] | List[List[List[float]]]): the data to be fitted
        rb_identifier (str): the RB identifier, either "stdrb", "irb" or "mrb"
        simultaneous_fit_vars (List[str], optional): the list of variables used to fit simultaneously
    Returns:
        A tuple of fitting data (list of lists of average fidelities or polarizations) and MRB fit parameters
    """
    n_qubits = len(qubit_set)

    fidelity_guess = 0.988**n_qubits
    offset_guess = 0.2 if n_qubits == 2 else 0.65
    amplitude_guess = 0.8 if n_qubits == 2 else 0.35

    estimates = {
        "depolarization_probability": 2 * (1 - fidelity_guess),
        "offset": offset_guess,
        "amplitude": amplitude_guess + offset_guess,
    }
    constraints = {
        "depolarization_probability": {"min": 0, "max": 1},
        "offset": {"min": 0, "max": 1},
        "amplitude": {"min": 0, "max": 1},
    }
    if rb_identifier in ("clifford", "mrb"):
        fit_data = np.array([np.mean(data, axis=1)])
        if rb_identifier == "clifford":
            params = create_multi_dataset_params(
                func, fit_data, initial_guesses=estimates, constraints=None, simultaneously_fit_vars=None
            )
            params.add(f"p_rb", expr=f"1-depolarization_probability_{1}")
            params.add(f"fidelity_per_clifford", expr=f"p_rb + (1 - p_rb) / (2**{n_qubits})")
        else:
            params = create_multi_dataset_params(
                func, fit_data, initial_guesses=None, constraints=constraints, simultaneously_fit_vars=None
            )
            params.add(f"p_mrb", expr=f"1-depolarization_probability_{1}")
            params.add(f"fidelity_mrb", expr=f"1 - (1 - p_mrb) * (1 - 1 / (4 ** {n_qubits}))")
    else:
        fit_data = np.array([np.mean(data[0], axis=1), np.mean(data[1], axis=1)])
        params = create_multi_dataset_params(
            func,
            fit_data,
            initial_guesses=estimates,
            constraints=None,
            simultaneously_fit_vars=simultaneous_fit_vars,
        )
        params.add(f"p_rb", expr=f"1-depolarization_probability_{1}")
        params.add(f"fidelity_per_clifford", expr=f"p_rb + (1 - p_rb) / (2**{n_qubits})")
        params.add(f"p_irb", expr=f"1-depolarization_probability_{2}")
        params.add(f"interleaved_fidelity", expr=f"p_irb / p_rb + (1 - p_irb / p_rb) / (2**{n_qubits})")

    return fit_data, params


@timeit
def generate_all_rb_circuits(
    qubits: List[int],
    sequence_lengths: List[int],
    clifford_dict: Dict[str, QuantumCircuit],
    num_circuit_samples: int,
    backend_arg: str | IQMBackendBase,
    interleaved_gate: Optional[QuantumCircuit],
) -> Tuple[Dict[int, List[QuantumCircuit]], Dict[int, List[QuantumCircuit]]]:
    """
    Args:
        qubits (List[int]): List of qubits
        sequence_lengths (List[int]): List of sequence lengths
        clifford_dict (Dict[str, QuantumCircuit]): the dictionary of Clifford circuits
        num_circuit_samples (int): the number of circuits samples
        backend_arg (str | IQMBackendBase): the backend fir which to generate the circuits.
        interleaved_gate (str): the name of the interleaved gate
    Returns:
        Tuple of untranspiled and transpiled circuits for all class-defined sequence lengths
    """
    untranspiled = {}
    transpiled = {}
    for s in sequence_lengths:
        qcvv_logger.info(f"Now at sequence length {s}")
        untranspiled[s], transpiled[s] = generate_random_clifford_seq_circuits(
            qubits, clifford_dict, s, num_circuit_samples, backend_arg, interleaved_gate
        )
    return untranspiled, transpiled


# pylint: disable=too-many-branches, disable=too-many-statements
@timeit
def generate_fixed_depth_parallel_rb_circuits(
    qubits_array: List[List[int]],
    cliffords_1q: Dict[str, QuantumCircuit],
    cliffords_2q: Dict[str, QuantumCircuit],
    sequence_length: int,
    num_samples: int,
    backend_arg: IQMBackendBase | str,
    interleaved_gate: Optional[QuantumCircuit] = None,
) -> Tuple[List[QuantumCircuit], List[QuantumCircuit]]:
    """Generates parallel RB circuits, before and after transpilation, at fixed depth

    Args:
        qubits_array (List[List[int]]): the qubits entering the quantum circuits
        cliffords_1q (Dict[str, QuantumCircuit]): dictionary of 1-qubit Cliffords in terms of IQM-native r and CZ gates
        cliffords_2q (Dict[str, QuantumCircuit]): dictionary of 2-qubit Cliffords in terms of IQM-native r and CZ gates
        sequence_length (int): the number of random Cliffords in the circuits
        num_samples (int): the number of circuit samples
        backend_arg (IQMBackendBase | str): the backend to transpile the circuits to
        interleaved_gate (Optional[QuantumCircuit]): whether the circuits should have interleaved gates
    Returns:
        A list of QuantumCircuits of given RB sequence length for parallel RB
    """

    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    # Identify total amount of qubits
    qubit_counts = [len(x) for x in qubits_array]

    # Shuffle qubits_array: we don't want unnecessary qubit registers
    shuffled_qubits_array = relabel_qubits_array_from_zero(qubits_array)
    # The total amount of qubits the circuits will have
    n_qubits = sum(qubit_counts)

    # Get the keys of the Clifford dictionaries
    clifford_1q_keys = list(cliffords_1q.keys())
    clifford_2q_keys = list(cliffords_2q.keys())

    # Generate the circuit samples
    circuits_list: List[QuantumCircuit] = []
    circuits_transpiled_list: List[QuantumCircuit] = []
    for _ in range(num_samples):
        circuit = QuantumCircuit(n_qubits)
        # Need to track inverses on each qubit, or each pair of qubits, separately
        circ_inverses = [QuantumCircuit(n) for n in qubit_counts]
        # Place Clifford sequences on each
        for _ in range(sequence_length):
            # Sample the random Cliffords for each qubit or qubits
            cliffords = []
            for n in qubit_counts:
                if n == 1:
                    rand_key = random.choice(clifford_1q_keys)
                    c_1q = cast(dict, cliffords_1q)[rand_key]
                    cliffords.append(c_1q)
                else:
                    rand_key = random.choice(clifford_2q_keys)
                    c_2q = cast(dict, cliffords_2q)[rand_key]
                    cliffords.append(c_2q)
            # Compose the sampled Cliffords in the circuit and add barrier
            for n, shuffled_qubits in enumerate(shuffled_qubits_array):
                circuit.compose(cliffords[n], qubits=shuffled_qubits, inplace=True)
                circ_inverses[n].compose(cliffords[n], qubits=list(range(qubit_counts[n])), inplace=True)
            circuit.barrier()  # NB: mind the barriers!

            if interleaved_gate is not None:
                for n, shuffled_qubits in enumerate(shuffled_qubits_array):
                    circuit.compose(interleaved_gate, qubits=shuffled_qubits, inplace=True)
                    circ_inverses[n].compose(interleaved_gate, qubits=list(range(qubit_counts[n])), inplace=True)
                circuit.barrier()

        # Compile and compose the inverse
        # compiled_inverse_clifford = []
        for n, shuffled_qubits in enumerate(shuffled_qubits_array):
            clifford_dict = cliffords_1q if qubit_counts[n] == 1 else cliffords_2q
            circuit.compose(
                compute_inverse_clifford(circ_inverses[n], clifford_dict),
                qubits=shuffled_qubits,
                inplace=True,
            )
        # Add measurement
        circuit.measure_all()

        circuits_list.append(circuit)
        # Transpilation here only means to the layout - avoid using automated transpilers!
        circuit_transpiled = QuantumCircuit(backend.num_qubits)
        qubits_array_flat = [x for y in qubits_array for x in y]
        circuit_transpiled.compose(circuit, qubits=qubits_array_flat, inplace=True)
        circuits_transpiled_list.append(circuit_transpiled)

    return circuits_list, circuits_transpiled_list


def generate_random_clifford_seq_circuits(
    qubits: List[int],
    clifford_dict: Dict[str, QuantumCircuit],
    seq_length: int,
    num_circ_samples: int,
    backend_arg: str | IQMBackendBase,
    interleaved_gate: Optional[QuantumCircuit] = None,
) -> Tuple[List[QuantumCircuit], List[QuantumCircuit]]:
    """Generate random Clifford circuits in native gates for a given sequence length.

    Args:
        qubits (List[int]): the list of qubits
        clifford_dict (Dict[str, QuantumCircuit]): A dictionary of Clifford gates labeled by (de)stabilizers
        seq_length (int): the sequence length
        num_circ_samples (int): the number of samples
        backend_arg (str | IQMBackendBase):
        interleaved_gate (Optional[QuantumCircuit]): Clifford native gate to be interleaved - None by default
    Returns:
        List[QuantumCircuit]: the list of `self.num_samples` random Clifford quantum circuits
    """
    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    qc_list = []  # List of random Clifford circuits
    qc_list_transpiled = []
    num_qubits = len(qubits)
    logic_qubits = list(range(num_qubits))

    if num_qubits > 2:
        raise ValueError("Please specify qubit layouts with only n=1 or n=2 qubits. Run MRB for n>2 instead.")

    clifford_keys = []
    if clifford_dict is not None:
        clifford_keys = list(clifford_dict.keys())

    for _ in range(num_circ_samples):
        qc = QuantumCircuit(num_qubits)
        qc_inv = QuantumCircuit(num_qubits)
        # Compose circuit with random Clifford gates
        for _ in range(seq_length):
            rand_key = random.choice(clifford_keys)
            clifford = cast(dict, clifford_dict)[rand_key]
            # Append clifford
            qc.compose(clifford, qubits=logic_qubits, inplace=True)
            # Append clifford gate to circuit to invert
            qc_inv.compose(clifford, qubits=logic_qubits, inplace=True)
            qc.barrier()
            # Append interleaved if not None
            if interleaved_gate is not None:
                qc.compose(interleaved_gate, qubits=logic_qubits, inplace=True)
                qc_inv.compose(interleaved_gate, qubits=logic_qubits, inplace=True)
                qc.barrier()
        # Append the inverse
        qc.compose(
            compute_inverse_clifford(qc_inv, clifford_dict),
            qubits=logic_qubits,
            inplace=True,
        )
        qc.measure_all()

        qc_list.append(qc)
        qc_transpiled = QuantumCircuit(backend.num_qubits)
        qc_transpiled.compose(qc, qubits=qubits, inplace=True)
        qc_list_transpiled.append(qc_transpiled)

    return qc_list, qc_list_transpiled


def get_survival_probabilities(num_qubits: int, counts: List[Dict[str, int]]) -> List[float]:
    """Compute a result's probability of being on the ground state.

    Args:
        num_qubits (int): the number of qubits
        counts (List[Dict[str, int]]): the result of the execution of a list of quantum circuits (counts)

    Returns:
        List[float]: the ground state probabilities of the RB sequence
    """
    return [c["0" * num_qubits] / sum(c.values()) if "0" * num_qubits in c.keys() else 0 for c in counts]


def import_native_gate_cliffords() -> Tuple[Dict[str, QuantumCircuit], Dict[str, QuantumCircuit]]:
    """Import native gate Clifford dictionaries
    Returns:
        Dictionaries of 1Q and 2Q Clifford gates
    """
    # Import the native-gate Cliffords
    with open(os.path.join(os.path.dirname(__file__), "clifford_1q.pkl"), "rb") as f1q:
        clifford_1q_dict = pickle.load(f1q)
    with open(os.path.join(os.path.dirname(__file__), "clifford_2q.pkl"), "rb") as f2q:
        clifford_2q_dict = pickle.load(f2q)
    qcvv_logger.info("Clifford dictionaries imported successfully !")
    return clifford_1q_dict, clifford_2q_dict


def lmfit_minimizer(
    fit_parameters: Parameters, fit_data: np.ndarray, depths: List[int], func: Callable
) -> MinimizerResult:
    """
    Args:
        fit_parameters (Parameters): the parameters to fit
        fit_data (np.ndarray): the data to fit
        depths (List[int]): the depths of the RB experiment
        func (Callable): the model function for fitting
    Returns:
        MinimizerResult: the result of the minimization
    """
    return minimize(
        fcn=multi_dataset_residual,
        params=fit_parameters,
        args=(depths, fit_data, func),
    )


def relabel_qubits_array_from_zero(arr: List[List[int]]) -> List[List[int]]:
    """Helper function to relabel a qubits array to an increasingly ordered one starting from zero
    e.g., [[2,3], [5], [7,8]]  ->  [[0,1], [2], [3,4]]
    Note: this assumes the input array is sorted in increasing order!
    """
    # Flatten the original array
    flat_list = [item for sublist in arr for item in sublist]
    # Generate a list of ordered numbers with the same length as the flattened array
    ordered_indices = list(range(len(flat_list)))
    # Reconstruct the list of lists structure
    result = []
    index = 0
    for sublist in arr:
        result.append(ordered_indices[index : index + len(sublist)])
        index += len(sublist)
    return result


def submit_parallel_rb_job(
    backend_arg: IQMBackendBase,
    qubits_array: List[List[int]],
    depth: int,
    sorted_transpiled_circuit_dicts: Dict[Tuple[int, ...], List[QuantumCircuit]],
    shots: int,
    calset_id: Optional[str],
    max_gates_per_batch: Optional[str],
) -> Dict[str, Any]:
    """Submit fixed-depth parallel MRB jobs for execution in the specified IQMBackend
    Args:
        backend_arg (IQMBackendBase): the IQM backend to submit the job
        qubits_array (List[int]): the qubits to identify the submitted job
        depth (int): the depth (number of canonical layers) of the circuits to identify the submitted job
        sorted_transpiled_circuit_dicts (Dict[Tuple[int,...], List[QuantumCircuit]]): A dictionary containing all MRB circuits
        shots (int): the number of shots to submit the job
        calset_id (Optional[str]): the calibration identifier
        max_gates_per_batch (Optional[str]): the maximum number of gates per batch to submit the job
    Returns:
        Dict with qubit layout, submitted job objects, type (vanilla/DD) and submission time
    """
    # Submit
    # Send to execute on backend
    # pylint: disable=unbalanced-tuple-unpacking
    execution_jobs, time_submit = submit_execute(
        sorted_transpiled_circuit_dicts, backend_arg, shots, calset_id, max_gates_per_batch=max_gates_per_batch
    )
    rb_submit_results = {
        "qubits": qubits_array,
        "depth": depth,
        "jobs": execution_jobs,
        "time_submit": time_submit,
    }
    return rb_submit_results


def submit_sequential_rb_jobs(
    qubits: List[int],
    transpiled_circuits: Dict[int, List[QuantumCircuit]],
    shots: int,
    backend_arg: str | IQMBackendBase,
    calset_id: Optional[str] = None,
    max_gates_per_batch: Optional[int] = None,
) -> List[Dict[str, Any]]:
    """Submit sequential RB jobs for execution in the specified IQMBackend
    Args:
        qubits (List[int]): the qubits to identify the submitted job
        transpiled_circuits (Dict[str, List[QuantumCircuit]]): A dictionary containing all MRB circuits
        shots (int): the number of shots to submit per job
        backend_arg (IQMBackendBase): the IQM backend to submit the job
        calset_id (Optional[str]): the calibration identifier
        max_gates_per_batch (Optional[int]): the maximum number of gates per batch
    Returns:
        Dict with qubit layout, submitted job objects, type (vanilla/DD) and submission time
    """
    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    all_submit_results = []
    rb_submit_results = {}
    for depth in transpiled_circuits.keys():
        # Submit - send to execute on backend
        # pylint: disable=unbalanced-tuple-unpacking
        execution_jobs, time_submit = submit_execute(
            {tuple(qubits): transpiled_circuits[depth]}, backend, shots, calset_id, max_gates_per_batch
        )
        rb_submit_results[depth] = {
            "qubits": qubits,
            "depth": depth,
            "jobs": execution_jobs,
            "time_submit": time_submit,
        }
        all_submit_results.append(rb_submit_results[depth])

    return all_submit_results


def survival_probabilities_parallel(
    qubits_array: List[List[int]], counts: List[Dict[str, int]]
) -> Dict[str, List[float]]:
    """Estimates marginalized survival probabilities from a parallel RB execution (at fixed depth)
    Args:
        qubits_array (List[int]): List of qubits in which the experiment was performed
        counts (Dict[str, int]): The measurement counts for corresponding bitstrings
    Returns:
        Dict[str, List[float]]: The survival probabilities for each qubit
    """
    # Global probability estimations
    global_probabilities = [{k: v / sum(c.values()) for k, v in c.items()} for c in counts]

    # The bit indices
    all_bit_indices = relabel_qubits_array_from_zero(qubits_array)

    # Estimate all marginal probabilities
    marginal_probabilities: Dict[str, List[Dict[str, float]]] = {str(q): [] for q in qubits_array}
    for position, indices in enumerate(all_bit_indices):
        marginal_probabilities[str(qubits_array[position])] = [
            marginal_distribution(global_probability, indices) for global_probability in global_probabilities
        ]

    # Estimate the survival probabilities in the marginal distributions
    marginal_survival_probabilities: Dict[str, List[float]] = {str(q): [] for q in qubits_array}
    for q in qubits_array:
        n_qubits = len(q)
        marginal_survival_probabilities[str(q)] = [
            c["0" * n_qubits] / sum(c.values()) if "0" * n_qubits in c.keys() else 0
            for c in marginal_probabilities[str(q)]
        ]

    return marginal_survival_probabilities


# pylint: disable=too-many-statements
def plot_rb_decay(
    identifier: str,
    qubits_array: List[List[int]],
    dataset: xr.Dataset,
    observations: Dict[int, Dict[str, Any]],
    violin: bool = True,
    scatter: bool = True,
    bars: bool = False,
    shade_stdev: bool = False,
    shade_meanerror: bool = False,
    logscale: bool = True,
    interleaved_gate: Optional[str] = None,
    mrb_2q_density: Optional[float] = None,
    mrb_2q_ensemble: Optional[Dict[str, float]] = None,
) -> Tuple[str, Figure]:
    """Plot the fidelity decay and the fit to the model.

    Args:
        identifier (str): the type of RB experiment
        qubits_array (List[List[int]]): Array of sets of qubits for which to plot decays
        dataset (xr.dataset): the dataset from the experiment
        observations (Dict[str, Dict[str, Any]]): the corresponding observations from the experiment
        bars (bool, optional): Whether error bars are plotted or not. Defaults to False
        violin (bool, optional): Whether violins are plotted or not. Defaults to True
        scatter (bool, optional): Whether all individual points are plotted or not. Defaults to True
        shade_stdev (bool, optional): Whether standard deviations are shaded or not. Defaults to False
        shade_meanerror (bool, optional): Whether to shade standard deviations. Defaults to False
        logscale (bool, optional): Whether x-axis uses logscale. Defaults to True
        interleaved_gate (Optional[str]):
        mrb_2q_density (Optional[float], optional): Density of MRB 2Q gates. Defaults to None.
        mrb_2q_ensemble (Optional[Dict[str, float]], optional): MRB ensemble of 2Q gates. Defaults to None.

    Returns:
        Tuple[str, Figure]: the plot title and the figure
    """
    fig, ax = plt.subplots()

    cmap = plt.get_cmap("winter")

    num_circuit_samples = dataset.attrs["num_circuit_samples"]
    timestamp = dataset.attrs["execution_timestamp"]
    backend_name = dataset.attrs["backend_name"]

    # Fetch the relevant observations indexed by qubit layouts
    depths = {}
    polarizations = {}
    average_polarizations = {}
    stddevs_from_mean = {}
    fidelity_value = {}
    fidelity_stderr = {}
    decay_rate = {}
    offset = {}
    amplitude = {}
    rb_type_keys = []
    colors = []
    if identifier != "irb":
        rb_type_keys = [identifier]
        colors = [cmap(i) for i in np.linspace(start=1, stop=0, num=len(qubits_array)).tolist()]
        if identifier == "mrb":
            depths[identifier] = {
                str(q): list(dataset.attrs[q_idx]["polarizations"].keys()) for q_idx, q in enumerate(qubits_array)
            }
            polarizations[identifier] = {
                str(q): dataset.attrs[q_idx]["polarizations"] for q_idx, q in enumerate(qubits_array)
            }
            average_polarizations[identifier] = {
                str(q): dataset.attrs[q_idx]["avg_polarization_nominal_values"] for q_idx, q in enumerate(qubits_array)
            }
            stddevs_from_mean[identifier] = {
                str(q): dataset.attrs[q_idx]["avg_polatization_stderr"] for q_idx, q in enumerate(qubits_array)
            }
        else:
            depths[identifier] = {
                str(q): list(dataset.attrs[q_idx]["fidelities"].keys()) for q_idx, q in enumerate(qubits_array)
            }
            polarizations[identifier] = {
                str(q): dataset.attrs[q_idx]["fidelities"] for q_idx, q in enumerate(qubits_array)
            }
            average_polarizations[identifier] = {
                str(q): dataset.attrs[q_idx]["avg_fidelities_nominal_values"] for q_idx, q in enumerate(qubits_array)
            }
            stddevs_from_mean[identifier] = {
                str(q): dataset.attrs[q_idx]["avg_fidelities_stderr"] for q_idx, q in enumerate(qubits_array)
            }
        # These are common to both MRB and standard Clifford
        fidelity_value[identifier] = {
            str(q): observations[q_idx]["avg_gate_fidelity"]["value"] for q_idx, q in enumerate(qubits_array)
        }
        fidelity_stderr[identifier] = {
            str(q): observations[q_idx]["avg_gate_fidelity"]["uncertainty"] for q_idx, q in enumerate(qubits_array)
        }
        decay_rate[identifier] = {
            str(q): dataset.attrs[q_idx]["decay_rate"]["value"] for q_idx, q in enumerate(qubits_array)
        }
        offset[identifier] = {
            str(q): dataset.attrs[q_idx]["fit_offset"]["value"] for q_idx, q in enumerate(qubits_array)
        }
        amplitude[identifier] = {
            str(q): dataset.attrs[q_idx]["fit_amplitude"]["value"] for q_idx, q in enumerate(qubits_array)
        }
    else:
        rb_type_keys = list(observations[0].keys())
        colors = [cmap(i) for i in np.linspace(start=1, stop=0, num=len(rb_type_keys)).tolist()]
        for rb_type in rb_type_keys:
            depths[rb_type] = {
                str(q): list(dataset.attrs[q_idx][rb_type]["fidelities"].keys()) for q_idx, q in enumerate(qubits_array)
            }
            polarizations[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["fidelities"] for q_idx, q in enumerate(qubits_array)
            }
            average_polarizations[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["avg_fidelities_nominal_values"]
                for q_idx, q in enumerate(qubits_array)
            }
            stddevs_from_mean[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["avg_fidelities_stderr"] for q_idx, q in enumerate(qubits_array)
            }
            fidelity_value[rb_type] = {
                str(q): observations[q_idx][rb_type]["avg_gate_fidelity"]["value"]
                for q_idx, q in enumerate(qubits_array)
            }
            fidelity_stderr[rb_type] = {
                str(q): observations[q_idx][rb_type]["avg_gate_fidelity"]["uncertainty"]
                for q_idx, q in enumerate(qubits_array)
            }
            decay_rate[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["decay_rate"]["value"] for q_idx, q in enumerate(qubits_array)
            }
            print(dataset.attrs)
            offset[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["fit_offset"]["value"] for q_idx, q in enumerate(qubits_array)
            }
            amplitude[rb_type] = {
                str(q): dataset.attrs[q_idx][rb_type]["fit_amplitude"]["value"] for q_idx, q in enumerate(qubits_array)
            }

    for index_irb, key in enumerate(rb_type_keys):
        for index_qubits, qubits in enumerate(qubits_array):
            # Index for colors
            index = index_irb if identifier == "irb" else index_qubits
            # Plot Averages
            # Draw the averages (w or w/o error bars)
            if bars:
                y_err_f = stddevs_from_mean[key][str(qubits)]
            else:
                y_err_f = None
            ax.errorbar(
                depths[key][str(qubits)],
                average_polarizations[key][str(qubits)].values(),
                yerr=y_err_f,
                # label=f"Avg with {num_circ_samples} samples",
                capsize=4,
                color=colors[index],
                fmt="o",
                mec="black",
                alpha=1,
                markersize=5,
            )
            if shade_stdev:
                # Shade the standard deviation
                ax.fill_between(
                    depths[key][str(qubits)],
                    [
                        a - np.sqrt(num_circuit_samples) * b
                        for a, b in zip(
                            average_polarizations[key][str(qubits)].values(),
                            stddevs_from_mean[key][str(qubits)].values(),
                        )
                    ],
                    [
                        a + np.sqrt(num_circuit_samples) * b
                        for a, b in zip(
                            average_polarizations[key][str(qubits)].values(),
                            stddevs_from_mean[key][str(qubits)].values(),
                        )
                    ],
                    color=colors[index],
                    alpha=0.2,
                    interpolate=True,
                )
            if shade_meanerror:
                # Shade the error from the mean
                ax.fill_between(
                    depths[key][str(qubits)],
                    [
                        a - b
                        for a, b in zip(
                            average_polarizations[key][str(qubits)].values(),
                            stddevs_from_mean[key][str(qubits)].values(),
                        )
                    ],
                    [
                        a + b
                        for a, b in zip(
                            average_polarizations[key][str(qubits)].values(),
                            stddevs_from_mean[key][str(qubits)].values(),
                        )
                    ],
                    color=colors[index],
                    alpha=0.1,
                    interpolate=True,
                )
            if violin:
                widths = [0.5 * m for m in depths[key][str(qubits)]]
                violin_parts_f = ax.violinplot(
                    polarizations[key][str(qubits)].values(),
                    positions=depths[key][str(qubits)],
                    showmeans=False,
                    showextrema=False,
                    widths=widths,
                )
                assert isinstance(violin_parts_f["bodies"], list)
                for pc in violin_parts_f["bodies"]:
                    assert isinstance(pc, PolyCollection)
                    pc.set_facecolor(colors[index])
                    pc.set_edgecolor("g")
                    pc.set_alpha(0.1)
            if scatter:
                # Plot the individual outcomes
                flat_fidelity_data = list(chain.from_iterable(polarizations[key][str(qubits)].values()))
                scatter_x = [
                    [depth] * len(polarizations[key][str(qubits)][depth]) for depth in depths[key][str(qubits)]
                ]
                flat_scatter_x = list(chain.from_iterable(scatter_x))
                ax.scatter(
                    flat_scatter_x,
                    flat_fidelity_data,
                    s=3,
                    color=colors[index],
                    alpha=0.35,
                )
            # Define the points of the x-axis
            x_linspace = np.linspace(np.min(depths[key][str(qubits)]), np.max(depths[key][str(qubits)]), 300)
            # Calculate the fit points of the y-axis
            y_fit = exponential_rb(
                x_linspace,
                1 - decay_rate[key][str(qubits)],
                offset[key][str(qubits)],
                amplitude[key][str(qubits)],
            )
            # Plot
            # Handle None types in instances where fidelity or stderr were not set
            if fidelity_value[key][str(qubits)] is None:
                fidelity_value[key][str(qubits)] = np.nan
            if fidelity_stderr[key][str(qubits)] is None:
                fidelity_stderr[key][str(qubits)] = np.nan

            if identifier == "mrb":
                plot_label = fr"$\overline{{F}}_\text{{MRB}} (n={len(qubits)})$ = {100.0 * fidelity_value[key][str(qubits)]:.2f} +/- {100.0 * fidelity_stderr[key][str(qubits)]:.2f} (%)"
            elif key == "interleaved":
                plot_label = fr"$\overline{{F}}_\text{{{interleaved_gate}}} ({qubits})$ = {100.0 * fidelity_value[key][str(qubits)]:.2f} +/- {100.0 * fidelity_stderr[key][str(qubits)]:.2f} (%)"
            else:
                plot_label = fr"$\overline{{F}}_\text{{Clifford}} ({qubits})$ = {100.0 * fidelity_value[key][str(qubits)]:.2f} +/- {100.0 * fidelity_stderr[key][str(qubits)]:.2f} (%)"

            ax.plot(
                x_linspace,
                y_fit,
                color=colors[index],
                alpha=0.5,
                label=plot_label,
            )

    # Labels
    on_qubits = "on all qubit layouts"
    fig_name = "all_qubit_layouts"
    if len(qubits_array) == 1:
        on_qubits = f"on qubits {qubits_array[0]}"
        fig_name = f"{str(qubits_array[0])}"

    if identifier == "irb":
        ax.set_title(
            f"{identifier.upper()} experiment for {interleaved_gate} {on_qubits}\nbackend: {backend_name} --- {timestamp}"
        )
    elif identifier == "clifford":
        ax.set_title(f"{identifier.capitalize()} experiment {on_qubits}\nbackend: {backend_name} --- {timestamp}")
    else:
        ax.set_title(
            f"{identifier.upper()} experiment {on_qubits}\n"
            f"2Q gate density: {mrb_2q_density}, ensemble {mrb_2q_ensemble}\n"
            f"backend: {backend_name} --- {timestamp}"
        )
    if identifier == "mrb":
        ax.set_ylabel("Polarization")
        ax.set_xlabel("Layer Depth")
    else:
        ax.set_ylabel("Fidelity")
        ax.set_xlabel("Sequence Length")
    if logscale:
        ax.set_xscale("log")

    # Show the relevant depths!
    # Gather and flatten all the possible depths and then pick unique elements
    all_depths = [x for d in depths.values() for w in d.values() for x in w]

    xticks = sorted(list(set(all_depths)))
    ax.set_xticks(xticks, labels=xticks)

    plt.legend(fontsize=8)
    ax.grid()

    plt.gcf().set_dpi(250)

    plt.close()

    return fig_name, fig


def validate_rb_qubits(qubits_array: List[List[int]], backend_arg: str | IQMBackendBase):
    """Validate qubit inputs for a Clifford RB experiment
    Args:
        qubits_array (List[List[int]]): the array of qubits
        backend_arg (IQMBackendBase): the IQM backend
    Raises:
        ValueError if specified pairs of qubits are not connected
    """
    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg
    # Identify total amount of qubits
    qubit_counts = [len(x) for x in qubits_array]

    # Validations
    # Suggest MRB for n>2 qubits
    if any(n > 2 for n in qubit_counts):
        raise ValueError("Please specify qubit layouts with only n=1 or n=2 qubits. Run MRB for n>2 instead.")
    pairs_qubits = [qubits_array[i] for i, n in enumerate(qubit_counts) if n == 2]
    # Qubits should be connected
    if any(tuple(x) not in backend.coupling_map for x in pairs_qubits):
        raise ValueError("Some specified pairs of qubits are not connected")


def validate_irb_gate(
    gate_id: str,
    backend_arg: str | IQMBackendBase,
    gate_params: Optional[List[float]] = None,
) -> QuantumCircuit:
    """Validate that an input gate is Clifford and transpiled to IQM's native basis

    Args:
        gate_id (str): the gate identifier as a Qiskit circuit library operator
        backend_arg (IQMBackendBase): the IQM backend to verify transpilation
        gate_params (Optional[List[float]]): the gate parameters
    Returns:
        Transpiled circuit

    """
    if gate_params is not None:
        gate = getattr(import_module("qiskit.circuit.library"), gate_id)(*gate_params)
    else:
        gate = getattr(import_module("qiskit.circuit.library"), gate_id)()

    if isinstance(backend_arg, str):
        backend = get_iqm_backend(backend_arg)
    else:
        backend = backend_arg

    qc = QuantumCircuit(gate.num_qubits)
    if gate in backend.operation_names:
        qc.compose(gate, qubits=range(gate.num_qubits), inplace=True)
        return qc

    # else transpile
    qc.compose(gate, qubits=range(gate.num_qubits), inplace=True)
    # Use Optimize SQG but retain final r gate - want the unitary to be preserved!
    qc_t = optimize_single_qubit_gates(
        transpile(qc, basis_gates=backend.operation_names, optimization_level=3), drop_final_rz=False
    )
    return qc_t



# Copyright 2024 IQM Benchmarks developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""AUTOMATED WEEKLY EXPERIMENT CONFIGURATION EXAMPLE
"""

from iqm.benchmarks.randomized_benchmarking.clifford_rb.clifford_rb import CliffordRBConfiguration
from iqm.benchmarks.randomized_benchmarking.interleaved_rb.interleaved_rb import InterleavedRBConfiguration
from iqm.benchmarks.randomized_benchmarking.mirror_rb.mirror_rb import MirrorRBConfiguration
from iqm.benchmarks.quantum_volume.clops import CLOPSConfiguration
from iqm.benchmarks.quantum_volume.quantum_volume import QuantumVolumeConfiguration
from iqm.benchmarks.entanglement.ghz import GHZConfiguration

WEEKLY_CRB = CliffordRBConfiguration(
    qubits_array=[[0,2]],
    sequence_lengths=[2**(m+1)-1 for m in range(7)],
    num_circuit_samples=25,
    shots=2**8,
    calset_id=None,
    parallel_execution=False,
)

WEEKLY_IRB = InterleavedRBConfiguration(
    qubits_array=[[0,2]],
    sequence_lengths=[2**(m+1)-1 for m in range(7)],
    num_circuit_samples=25,
    shots=2**8,
    calset_id=None,
    parallel_execution=False,
    interleaved_gate = "CZGate",
    interleaved_gate_params = None,
    simultaneous_fit = ["amplitude", "offset"],
)

WEEKLY_MRB = MirrorRBConfiguration(
    qubits_array=[[0,2],
                  [0,1,2],
                  [0,1,2,3],
                  [0,1,2,3,4]],
    depths_array=[[2**m for m in range(7)]],
    num_circuit_samples=10,
    num_pauli_samples=5,
    shots=2**8,
    qiskit_optim_level=1,
    routing_method="sabre",
    two_qubit_gate_ensemble={"CZGate": 1.0},
    density_2q_gates=0.35,
    calset_id=None,
)

WEEKLY_QV = QuantumVolumeConfiguration(
    num_circuits=500,
    shots=2**8,
    calset_id=None,
    num_sigmas=2,
    choose_qubits_routine="custom",
    custom_qubits_array=[[0,1,2,3,4]],
    #dynamical_decoupling=True,  # Default is False
    #dd_strategy=dd_custom,  # Default is dd.DDStrategy()
    qiskit_optim_level=3,
    optimize_sqg=True,
    routing_method="sabre",
    physical_layout="fixed",
    max_gates_per_batch=60_000,
    rem=True,
    mit_shots=1_000,
)

WEEKLY_CLOPS = CLOPSConfiguration(
    qubits=[0,1,2,3,4],
    num_circuits=100, # By definition set to 100
    num_updates=10, # By definition set to 10
    num_shots=100, # By definition set to 100
    calset_id=None,
    clops_h_bool=True,
    qiskit_optim_level=3,
    optimize_sqg=True,
    routing_method="sabre",
    physical_layout="fixed",
)

WEEKLY_GHZ = GHZConfiguration(
    state_generation_routine="tree",  # "naive", "log_depth" (default)
    custom_qubits_array=[[0,1,2,3,4]],
    shots=2**10,
    qiskit_optim_level=3,
    optimize_sqg=True,
    fidelity_routine="coherences",
    rem=True,
    mit_shots=1000,
)
